# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Tuple
from unittest import TestCase

from pyshimmer.bluetooth.bt_serial import BluetoothSerial
from pyshimmer.test_util import MockSerial


class BluetoothSerialTest(TestCase):

    @staticmethod
    def create_sot() -> Tuple[MockSerial, BluetoothSerial]:
        mock = MockSerial()
        # noinspection PyTypeChecker
        serial = BluetoothSerial(mock)
        return mock, serial

    def test_read_varlen(self):
        mock, sot = self.create_sot()

        mock.test_put_read_data(b'\x00\x04\x01\x02\x03\x04')
        r = sot.read_varlen()
        self.assertEqual(r, b'')

        r = sot.read_varlen()
        self.assertEqual(r, b'\x01\x02\x03\x04')

    def test_write_varlen(self):
        mock, sot = self.create_sot()

        sot.write_varlen(b'thisisatest')
        r = mock.test_get_write_data()

        self.assertEqual(r[0], 11)
        self.assertEqual(r[1:], b'thisisatest')

        self.assertRaises(ValueError, sot.write_varlen, b'A' * 300)

    def test_write_command(self):
        mock, sot = self.create_sot()

        sot.write_command(0x10, '>BH', 0x42, 0x10)
        r = mock.test_get_write_data()
        self.assertEqual(r, b'\x10\x42\x00\x10')

        sot.write_command(0x11, 'varlen', b'hello')
        r = mock.test_get_write_data()
        self.assertEqual(r, b'\x11\x05hello')

    def test_read_ack(self):
        mock, sot = self.create_sot()
        mock.test_put_read_data(b'\xFF\x00')

        sot.read_ack()
        self.assertRaises(ValueError, sot.read_ack)

    def test_read_response(self):
        mock, sot = self.create_sot()

        # int response code
        mock.test_put_read_data(b'\x42')
        r = sot.read_response(0x42)
        self.assertEqual(r, ())

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42')
        self.assertRaises(RuntimeError, sot.read_response, 0x43)

        # Single-byte tuple response code
        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42')
        r = sot.read_response((0x42,))
        self.assertEqual(r, ())

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42')
        self.assertRaises(RuntimeError, sot.read_response, (0x43,))

        # Multi-byte tuple response code
        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42\x43')
        r = sot.read_response((0x42, 0x43))
        self.assertEqual(r, ())

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x43\x42')
        self.assertRaises(RuntimeError, sot.read_response, (0x43, 0x44))

        # bytes response code
        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42\x43')
        r = sot.read_response(b'\x42\x43')
        self.assertEqual(r, ())

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x42\x44')
        self.assertRaises(RuntimeError, sot.read_response, b'\x43\x44')

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x50\x03\x00\x01\x02')
        r = sot.read_response(0x50, arg_format='varlen')
        self.assertEqual(r, b'\x00\x01\x02')

        mock.test_clear_read_buffer()
        mock.test_put_read_data(b'\x50\x03\x00\x01\x02')
        r = sot.read_response(0x50, arg_format='<HBB')
        self.assertEqual(r, (0x3, 0x1, 0x2))



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase

from pyshimmer.dev.channels import ChDataTypeAssignment, get_ch_dtypes, SensorChannelAssignment, SensorBitAssignments, \
    ChannelDataType, EChannelType, ESensorGroup, sort_sensors


class DeviceChannelsTest(TestCase):

    def test_channel_enum_uniqueness(self):
        try:
            # The exception will trigger upon import if the enum values are not unique
            from pyshimmer.dev.channels import EChannelType
        except ValueError as e:
            self.fail(f'Enum not unique: {e}')

    def test_channel_data_type_decoding(self):
        def test_both_endianess(byte_val_le: bytes, expected: int, signed: bool):
            blen = len(byte_val_le)
            dt_le = ChannelDataType(blen, signed=signed, le=True)
            dt_be = ChannelDataType(blen, signed=signed, le=False)

            self.assertEqual(expected, dt_le.decode(byte_val_le))
            self.assertEqual(expected, dt_be.decode(byte_val_le[::-1]))

        # Test the property getters
        dt = ChannelDataType(3, signed=False, le=True)
        self.assertEqual(dt.little_endian, True)
        self.assertEqual(dt.big_endian, False)
        self.assertEqual(dt.signed, False)
        self.assertEqual(dt.size, 3)

        # Test the property getters
        dt = ChannelDataType(3, signed=False, le=False)
        self.assertEqual(dt.little_endian, False)
        self.assertEqual(dt.big_endian, True)

        # Test unsigned decodation for 3 byte data
        test_both_endianess(b'\x00\x00\x00', 0x000000, signed=False)
        test_both_endianess(b'\x10\x00\x00', 0x000010, signed=False)
        test_both_endianess(b'\x00\x00\xFF', 0xFF0000, signed=False)
        test_both_endianess(b'\xFF\xFF\xFF', 0xFFFFFF, signed=False)

        # Test signed decodation for 3 byte data
        test_both_endianess(b'\xFF\xFF\xFF', -1, signed=True)
        test_both_endianess(b'\x00\x00\x80', -2 ** 23, signed=True)
        test_both_endianess(b'\xFF\xFF\x7F', 2 ** 23 - 1, signed=True)
        test_both_endianess(b'\xFF\x00\x00', 255, signed=True)

        # Test unsigned decodation for 2 byte data
        test_both_endianess(b'\x00\x00', 0x0000, signed=False)
        test_both_endianess(b'\x10\x00', 0x0010, signed=False)
        test_both_endianess(b'\x00\xFF', 0xFF00, signed=False)
        test_both_endianess(b'\xFF\xFF', 0xFFFF, signed=False)

        # Test signed decodation for 2 byte data
        test_both_endianess(b'\xFF\xFF', -1, signed=True)
        test_both_endianess(b'\x00\x80', -2 ** 15, signed=True)
        test_both_endianess(b'\xFF\x7F', 2 ** 15 - 1, signed=True)
        test_both_endianess(b'\xFF\x00', 255, signed=True)

    def test_channel_data_type_encoding(self):
        def test_both_endianess(val: int, val_len: int, expected: bytes, signed: bool):
            dt_le = ChannelDataType(val_len, signed=signed, le=True)
            dt_be = ChannelDataType(val_len, signed=signed, le=False)

            self.assertEqual(expected, dt_le.encode(val))
            self.assertEqual(expected[::-1], dt_be.encode(val))

        test_both_endianess(0x1234, 2, b'\x34\x12', signed=False)
        test_both_endianess(-0x10, 2, b'\xF0\xFF', signed=True)

        test_both_endianess(0x12345, 3, b'\x45\x23\x01', signed=False)
        test_both_endianess(-0x12345, 3, b'\xbb\xdc\xfe', signed=True)

    def test_get_ch_dtypes(self):
        channels = [EChannelType.INTERNAL_ADC_13, EChannelType.GYRO_MPU9150_Y]
        r = get_ch_dtypes(channels)

        self.assertEqual(len(r), 2)
        first, second = r

        self.assertEqual(first.size, 2)
        self.assertEqual(first.little_endian, True)
        self.assertEqual(first.signed, False)

        self.assertEqual(second.size, 2)
        self.assertEqual(second.little_endian, False)
        self.assertEqual(second.signed, True)

    def test_sensor_group_uniqueness(self):
        try:
            # The exception will trigger upon import if the enum values are not unique
            from pyshimmer.dev.channels import ESensorGroup
        except ValueError as e:
            self.fail(f'Enum not unique: {e}')

    def test_datatype_assignments(self):
        from pyshimmer.dev.channels import EChannelType
        for ch_type in EChannelType:
            if ch_type not in ChDataTypeAssignment:
                self.fail(f'No data type assigned to channel type: {ch_type}')

    def test_sensor_channel_assignments(self):
        from pyshimmer.dev.channels import ESensorGroup
        for sensor in ESensorGroup:
            if sensor not in SensorChannelAssignment:
                self.fail(f'No channels assigned to sensor type: {sensor}')

    def test_sensor_bit_assignments_uniqueness(self):
        for s1 in SensorBitAssignments.keys():
            for s2 in SensorBitAssignments.keys():
                if s1 != s2 and SensorBitAssignments[s1] == SensorBitAssignments[s2]:
                    self.fail(f'Colliding bitfield assignments for sensor {s1} and {s2}')

    def test_sort_sensors(self):
        sensors = [ESensorGroup.BATTERY, ESensorGroup.ACCEL_LN]
        expected = [ESensorGroup.ACCEL_LN, ESensorGroup.BATTERY]
        r = sort_sensors(sensors)
        self.assertEqual(r, expected)

        sensors = [ESensorGroup.CH_A15, ESensorGroup.MAG_MPU, ESensorGroup.ACCEL_LN, ESensorGroup.CH_A15]
        expected = [ESensorGroup.ACCEL_LN, ESensorGroup.CH_A15, ESensorGroup.CH_A15, ESensorGroup.MAG_MPU]
        r = sort_sensors(sensors)
        self.assertEqual(r, expected)



from serial import Serial

from pyshimmer import ShimmerDock, DEFAULT_BAUDRATE, fmt_hex


def main(args=None):    
        serial = Serial('/dev/ttyECGdev', DEFAULT_BAUDRATE)

        print(f'Connecting docker')
        shim_dock = ShimmerDock(serial)

        mac = shim_dock.get_mac_address()
        print(f'Device MAC: {fmt_hex(mac)}')

        shim_dock.close()


if __name__ == '__main__':
    main()        


# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from abc import ABC, abstractmethod
from typing import List, Tuple, Union, Iterable

from pyshimmer.bluetooth.bt_const import *
from pyshimmer.bluetooth.bt_serial import BluetoothSerial

from pyshimmer.dev.base import dr2sr, sr2dr, sec2ticks, ticks2sec
from pyshimmer.dev.channels import ChannelDataType, EChannelType, ESensorGroup, serialize_sensorlist
from pyshimmer.dev.exg import ExGRegister
from pyshimmer.dev.calibration import AllCalibration
from pyshimmer.dev.fw_version import get_firmware_type

from pyshimmer.util import bit_is_set, resp_code_to_bytes, calibrate_u12_adc_value, battery_voltage_to_percent


class DataPacket:
    """Parses data packets received by the Shimmer device

    :arg stream_types: List of tuples that contains each data channel contained in the data packet as well as the
        corresponding data type decoder
    """

    def __init__(self, stream_types: List[Tuple[EChannelType, ChannelDataType]]):
        self._types = stream_types
        self._values = {}

    @property
    def channels(self) -> List[EChannelType]:
        """The data channels present in this data packet

        :return: The channels as list
        """
        return [t for t, _ in self._types]

    @property
    def channel_types(self) -> List[ChannelDataType]:
        """The channel data types that represent the binary data of each channel

        :return: The data types as list
        """
        return [t for _, t in self._types]

    def __getitem__(self, item: EChannelType) -> any:
        """Return the value of a certain data channel

        :param item: The data channel for which to return the data
        :return: The value of the data channel
        """
        return self._values[item]

    def receive(self, ser: BluetoothSerial) -> None:
        """Receive and decode a data packet

        :param ser: The serial device from which to read the data
        """
        ser.read_response(DATA_PACKET)

        for channel_type, channel_dtype in self._types:
            data_bin = ser.read(channel_dtype.size)
            self._values[channel_type] = channel_dtype.decode(data_bin)


class ShimmerCommand(ABC):
    """Abstract base class that represents a command sent to the Shimmer

    """

    @abstractmethod
    def send(self, ser: BluetoothSerial) -> None:
        """Encodes the command and sends it to the Shimmer via the provided serial interface

        :param ser: The serial to use for sending the command
        """
        pass

    def has_response(self) -> bool:
        """Specifies if the command has a response that needs to be read from the return stream

        :return: True if the command has a response, else false
        """
        return False

    def get_response_code(self) -> bytes:
        """The response code of the command

        :return: The response code as a series of bytes, is normally one or two bytes long
        """
        return bytes()

    def receive(self, ser: BluetoothSerial) -> any:
        """Decode the command response from the provided serial interface

        :param ser: The serial from which to decode the response
        :return: The data contained in the response
        """
        return None


class ResponseCommand(ShimmerCommand, ABC):
    """Abstract base class for all commands that feature a command response

    :arg rcode: The response code of the response. Can be a single int for a single-byte response code or
        a tuple of ints or a bytes instance for a multi-byte response code
    """

    def __init__(self, rcode: Union[int, Tuple[int, ...], bytes]):
        self._rcode = resp_code_to_bytes(rcode)

    def has_response(self) -> bool:
        return True

    def get_response_code(self) -> bytes:
        return self._rcode


class OneShotCommand(ShimmerCommand):
    """Class for commands that only send a command code and have no response

    :arg cmd_code: The command code to send
    """

    def __init__(self, cmd_code: int):
        self._code = cmd_code

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(self._code)


class GetStringCommand(ResponseCommand):
    """Send a command that features a variable-length string as response

    :arg req_code: The command code of the request
    :arg resp_code: The response code
    :arg encoding: The encoding to use when reading the response string
    """

    def __init__(self, req_code: int, resp_code: Union[int, Tuple[int], bytes], encoding: str = 'utf8'):
        super().__init__(resp_code)
        self._req_code = req_code
        self._encoding = encoding

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(self._req_code)

    def receive(self, ser: BluetoothSerial) -> any:
        str_bin = ser.read_response(self._rcode, 'varlen')
        return str_bin.decode(self._encoding)


class SetStringCommand(ShimmerCommand):
    """A command for sending a variable-length string to the device

    :arg req_code: The code of the command request
    :arg str_data: The data to send as part of the request
    :arg encoding: The encoding to use when writing the data to the stream
    """

    def __init__(self, req_code: int, str_data: str, encoding: str = 'utf8'):
        self._req_code = req_code
        self._str_data = str_data
        self._encoding = encoding

    def send(self, ser: BluetoothSerial) -> None:
        str_bin = self._str_data.encode(self._encoding)
        ser.write_command(self._req_code, 'varlen', str_bin)


class GetSamplingRateCommand(ResponseCommand):
    """Retrieve the sampling rate in samples per second

    """

    def __init__(self):
        super().__init__(SAMPLING_RATE_RESPONSE)

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_SAMPLING_RATE_COMMAND)

    def receive(self, ser: BluetoothSerial) -> float:
        sr_clock = ser.read_response(SAMPLING_RATE_RESPONSE, arg_format='<H')
        sr = dr2sr(sr_clock)
        return sr


class SetSamplingRateCommand(ShimmerCommand):

    def __init__(self, sr: float):
        self._sr = sr

    def send(self, ser: BluetoothSerial) -> None:
        dr = sr2dr(self._sr)
        ser.write_command(SET_SAMPLING_RATE_COMMAND, "<H", dr)


class GetBatteryCommand(ResponseCommand):
    """Retrieve the battery state

    """

    def __init__(self, in_percent: bool):
        super().__init__(FULL_BATTERY_RESPONSE)
        self._in_percent = in_percent

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_BATTERY_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        batt = ser.read_response(self.get_response_code(), arg_format='BBB')
        # Calculation see:
        # http://shimmersensing.com/wp-content/docs/support/documentation/LogAndStream_for_Shimmer3_Firmware_User_Manual_rev0.11a.pdf (Page 17)
        # https://shimmersensing.com/wp-content/docs/support/documentation/Shimmer_User_Manual_rev3p.pdf (Page 53)
        raw_values = batt[1] * 256 + batt[0]
        batt_voltage = calibrate_u12_adc_value(raw_values, 0, 3.0, 1.0) * 1.988
        if self._in_percent:
            return battery_voltage_to_percent(batt_voltage)
        else:
            return batt_voltage


class GetConfigTimeCommand(ResponseCommand):
    """Retrieve the config time that is stored in the Shimmer device configuration file

    """

    def __init__(self):
        super().__init__(CONFIGTIME_RESPONSE)

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_CONFIGTIME_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        r = ser.read_response(CONFIGTIME_RESPONSE, arg_format='varlen')
        return int(r)


class SetConfigTimeCommand(ShimmerCommand):
    """Set the config time, which will be stored in the Shimmer device configuration file

    :arg time: The integer value to send
    """

    def __init__(self, time: int):
        self._time = time

    def send(self, ser: BluetoothSerial) -> None:
        time_str = '{:d}'.format(int(self._time))
        time_bin = time_str.encode('ascii')

        ser.write_command(SET_CONFIGTIME_COMMAND, "varlen", time_bin)


class GetRealTimeClockCommand(ResponseCommand):
    """
    Get the real-time clock as UNIX Timestamp in seconds
    """

    def __init__(self):
        super().__init__(RWC_RESPONSE)

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_RWC_COMMAND)

    def receive(self, ser: BluetoothSerial) -> float:
        t_ticks = ser.read_response(RWC_RESPONSE, arg_format="<Q")
        return ticks2sec(t_ticks)


class SetRealTimeClockCommand(ShimmerCommand):
    """
    Set the real-time clock as UNIX timestamp in seconds

    :arg ts_sec: The UNIX timestamp in seconds
    """

    def __init__(self, ts_sec: float):
        self._time = int(ts_sec)

    def send(self, ser: BluetoothSerial) -> None:
        t_ticks = sec2ticks(self._time)
        ser.write_command(SET_RWC_COMMAND, '<Q', t_ticks)


class GetStatusCommand(ResponseCommand):
    """Retrieve the current status of the device

    """
    STATUS_DOCKED_BF = 1 << 0
    STATUS_SENSING_BF = 1 << 1
    STATUS_RTC_SET_BF = 1 << 2
    STATUS_LOGGING_BF = 1 << 3
    STATUS_STREAMING_BF = 1 << 4
    STATUS_SD_PRESENT_BF = 1 << 5
    STATUS_SD_ERROR_BF = 1 << 6
    STATUS_RED_LED_BF = 1 << 7
    STATUS_BITFIELDS = (STATUS_DOCKED_BF, STATUS_SENSING_BF, STATUS_RTC_SET_BF, STATUS_LOGGING_BF, STATUS_STREAMING_BF,
                        STATUS_SD_PRESENT_BF, STATUS_SD_ERROR_BF, STATUS_RED_LED_BF)

    def __init__(self):
        super().__init__(FULL_STATUS_RESPONSE)

    def unpack_status_bitfields(self, val: int) -> List[bool]:
        values = [bit_is_set(val, f) for f in self.STATUS_BITFIELDS]
        return values

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_STATUS_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        bitfields = ser.read_response(self.get_response_code(), arg_format='B')
        return self.unpack_status_bitfields(bitfields)


class GetFirmwareVersionCommand(ResponseCommand):
    """Retrieve the firmware type and version

    """

    def __init__(self):
        super().__init__(FW_VERSION_RESPONSE)

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_FW_VERSION_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        fw_type_bin, major, minor, rel = ser.read_response(
            FW_VERSION_RESPONSE, arg_format='<HHBB')
        fw_type = get_firmware_type(fw_type_bin)
        return fw_type, major, minor, rel


class GetAllCalibrationCommand(ResponseCommand):
    """ Returns all the stored calibration values (84 bytes) in the following order:

            ESensorGroup.ACCEL_LN (21 bytes)
            ESensorGroup.GYRO     (21 bytes)
            ESensorGroup.MAG      (21 bytes)
            ESensorGroup.ACCEL_WR (21 bytes)        

        The breakdown of the kinematic (accel x 2, gyro and mag) calibration values is as follows:
            [bytes  0- 5] offset bias values: 3 (x,y,z) 16-bit signed integers (big endian). 
            [bytes  6-11] sensitivity values: 3 (x,y,z) 16-bit signed integers (big endian). 
            [bytes 12-20] alignment matrix:  9 values    8-bit signed integers.
    """

    def __init__(self):
        super().__init__(ALL_CALIBRATION_RESPONSE)

        self._offset = 0x0
        self._rlen = 0x54  # 84 bytes

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_ALL_CALIBRATION_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        ser.read_response(ALL_CALIBRATION_RESPONSE)
        reg_data = ser.read(self._rlen)
        return AllCalibration(reg_data)


class InquiryCommand(ResponseCommand):
    """Perform an inquiry to determine the sample rate, buffer size, and active data channels

    """

    def __init__(self):
        super().__init__(INQUIRY_RESPONSE)

    @staticmethod
    def decode_channel_types(ct_bin: bytes) -> List[EChannelType]:
        ctypes_index = struct.unpack('B' * len(ct_bin), ct_bin)
        ctypes = [BtChannelsByIndex[i] for i in ctypes_index]
        return ctypes

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(INQUIRY_COMMAND)

    def receive(self, ser: BluetoothSerial) -> any:
        sr_val, _, n_ch, buf_size = ser.read_response(
            INQUIRY_RESPONSE, arg_format='<HIBB')
        channel_conf = ser.read(n_ch)

        sr = dr2sr(sr_val)
        ctypes = self.decode_channel_types(channel_conf)

        return sr, buf_size, ctypes


class StartStreamingCommand(OneShotCommand):
    """Start streaming data over the Bluetooth channel

    """

    def __init__(self):
        super().__init__(START_STREAMING_COMMAND)


class StopStreamingCommand(OneShotCommand):
    """Stop streaming data over the Bluetooth channel

    """

    def __init__(self):
        super().__init__(STOP_STREAMING_COMMAND)


class GetEXGRegsCommand(ResponseCommand):
    """Retrieve the current state of the ExG chip register

    Queries the values of all registers of the specified chip and returns it as an ExGRegister instance

    :arg chip_id: The chip id, can be one of [0, 1]
    """

    def __init__(self, chip_id: int):
        super().__init__(EXG_REGS_RESPONSE)

        self._chip = chip_id
        self._offset = 0x0
        self._rlen = 0xA

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(GET_EXG_REGS_COMMAND, 'BBB',
                          self._chip, self._offset, self._rlen)

    def receive(self, ser: BluetoothSerial) -> any:
        rlen = ser.read_response(EXG_REGS_RESPONSE, arg_format='B')
        if not rlen == self._rlen:
            raise ValueError(
                'Response does not contain required amount of bytes')

        reg_data = ser.read(rlen)
        return ExGRegister(reg_data)


class SetEXGRegsCommand(ShimmerCommand):
    """Set the binary contents of the ExG registers of a chip

    :arg chip_id: The id of the chip, can be one of [0, 1]
    :arg offset: At which offset to write the data
    :arg data: The bytes to write to the registers
    """

    def __init__(self, chip_id: int, offset: int, data: bytes):
        self._chip = chip_id
        self._offset = offset
        self._data = data

    def send(self, ser: BluetoothSerial) -> None:
        dlen = len(self._data)
        ser.write_command(SET_EXG_REGS_COMMAND, 'BBB',
                          self._chip, self._offset, dlen)
        ser.write(self._data)


class GetExperimentIDCommand(GetStringCommand):
    """Retrieve the experiment id

    """

    def __init__(self):
        super().__init__(GET_EXPID_COMMAND, EXPID_RESPONSE)


class SetExperimentIDCommand(SetStringCommand):
    """Set the experiment id

    :arg exp_id: The experiment id as string
    """

    def __init__(self, exp_id: str):
        super().__init__(SET_EXPID_COMMAND, exp_id)


class SetSensorsCommand(ShimmerCommand):

    def __init__(self, sensors: Iterable[ESensorGroup]):
        self._sensors = list(sensors)

    def send(self, ser: BluetoothSerial) -> None:
        bitfield_bin = serialize_sensorlist(self._sensors)
        ser.write_command(SET_SENSORS_COMMAND, "<3s", bitfield_bin)


class GetDeviceNameCommand(GetStringCommand):
    """Get the device name

    """

    def __init__(self):
        super().__init__(GET_SHIMMERNAME_COMMAND, SHIMMERNAME_RESPONSE)


class SetDeviceNameCommand(SetStringCommand):
    """Set the device name

    :arg dev_name: The new device name as string
    """

    def __init__(self, dev_name: str):
        super().__init__(SET_SHIMMERNAME_COMMAND, dev_name)


class SetStatusAckCommand(ShimmerCommand):

    def __init__(self, enabled: bool):
        """Command to enable/disable the ACK byte before status messages

        By default, the Shimmer firmware sends an acknowledgment byte before
        sending unsolicited status messages to the host. This confuses the state
        machine of the Python API but is always expected by the official Shimmer
        software. This command is used by the Python API to automatically disable
        the acknowledgment when connecting to a Shimmer.

        :param enabled: If set to True, the acknowledgment is sent. If set to False,
            the acknowledgment is not sent.
        """
        self._enabled = enabled

    def send(self, ser: BluetoothSerial) -> None:
        ser.write_command(ENABLE_STATUS_ACK_COMMAND, "<B", int(self._enabled))


class StartLoggingCommand(OneShotCommand):
    """Begin logging data to the SD card

    """

    def __init__(self):
        super().__init__(START_LOGGING_COMMAND)


class StopLoggingCommand(OneShotCommand):
    """End logging data to the SD card

    """

    def __init__(self):
        super().__init__(STOP_LOGGING_COMMAND)


class DummyCommand(OneShotCommand):
    """Dummy command that is only acknowledged by the Shimmer but triggers no response

    """

    def __init__(self):
        super().__init__(DUMMY_COMMAND)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from queue import Queue, Empty
from threading import Event, Thread
from typing import List, Tuple, Callable, Iterable, Optional

from serial import Serial

from pyshimmer.bluetooth.bt_commands import ShimmerCommand, GetSamplingRateCommand, GetConfigTimeCommand, \
    SetConfigTimeCommand, GetRealTimeClockCommand, SetRealTimeClockCommand, GetStatusCommand, \
    GetFirmwareVersionCommand, InquiryCommand, StartStreamingCommand, StopStreamingCommand, DataPacket, \
    GetEXGRegsCommand, SetEXGRegsCommand, StartLoggingCommand, StopLoggingCommand, GetExperimentIDCommand, \
    SetExperimentIDCommand, GetDeviceNameCommand, SetDeviceNameCommand, DummyCommand, GetBatteryCommand, \
    SetSamplingRateCommand, SetSensorsCommand, SetStatusAckCommand, AllCalibration, GetAllCalibrationCommand
from pyshimmer.bluetooth.bt_const import ACK_COMMAND_PROCESSED, DATA_PACKET, FULL_STATUS_RESPONSE, INSTREAM_CMD_RESPONSE
from pyshimmer.bluetooth.bt_serial import BluetoothSerial
from pyshimmer.dev.channels import ChDataTypeAssignment, ChannelDataType, EChannelType, ESensorGroup
from pyshimmer.dev.exg import ExGRegister
from pyshimmer.dev.fw_version import EFirmwareType, FirmwareVersion, FirmwareCapabilities
from pyshimmer.serial_base import ReadAbort
from pyshimmer.util import fmt_hex, PeekQueue


class RequestCompletion:
    """
    Returned by the Bluetooth API upon sending a request. Signals the completion of a request when the API receives
    the corresponding acknowledgment.
    """

    def __init__(self):
        self.__event = Event()

    def set_completed(self) -> None:
        self.__event.set()

    def has_completed(self) -> bool:
        return self.__event.is_set()

    def wait(self) -> None:
        self.__event.wait()


class RequestResponse:
    """
    Returned by the Bluetooth API upon sending a request that features a response. Returns the request response data
    upon completion.
    """

    def __init__(self):
        self.__event = Event()
        self.__r = None

    def has_result(self) -> bool:
        return self.__event.is_set()

    def get_result(self) -> any:
        return self.__r

    def set_result(self, r: any) -> None:
        self.__r = r
        self.__event.set()

    def wait(self) -> any:
        self.__event.wait()
        return self.get_result()


class BluetoothRequestHandler:
    """Base class for the Bluetooth API which handles serial data processing synchronously

    In contrast to the :class:`ShimmerBluetooth` class which uses Threading, this class acts as a base layer that
    operates synchronously and allows for easier testing.

    :arg serial: The serial interface to use
    """

    def __init__(self, serial: BluetoothSerial):
        self._serial = serial

        self._ack_queue = Queue()
        self._resp_queue = PeekQueue()

        self._stream_types = []
        self._stream_cbs = []
        self._status_cbs = []

    def set_stream_types(self, types: List[Tuple[EChannelType, ChannelDataType]]) -> None:
        """Set the channel types that are streamed as part of the data packets

        :param types: A List of tuples, each containing a channel type and its corresponding data type
        """
        self._stream_types = types

    def add_stream_callback(self, cb: Callable[[DataPacket], None]) -> None:
        """Add a stream callback which is called when a new data packet arrives

        :param cb: a function with a single argument
        """
        self._stream_cbs += [cb]

    def remove_stream_callback(self, cb: Callable[[DataPacket], None]) -> None:
        """Remove the callback from the list of active callbacks

        :param cb: The callback function to remove
        """
        self._stream_cbs.remove(cb)

    def add_status_callback(self, cb: Callable[[List[bool]], None]) -> None:
        """Add a status callback which is called when a new status update from the Shimmer arrives

        :param cb: a function with a single argument. The argument is the same value is the return value of the
            :meth:`pyshimmer.bluetooth.bt_api.ShimmerBluetooth.get_status` method.
        """
        self._status_cbs += [cb]

    def remove_status_callback(self, cb: Callable[[List[bool]], None]) -> None:
        """Remove the callback from the list of active callbacks

        :param cb: The callback function to remove
        """
        self._status_cbs.remove(cb)

    def _process_ack(self):
        self._serial.read_ack()
        compl_obj, cmd_resp_pair = self._ack_queue.get_nowait()

        if None not in cmd_resp_pair:
            # We are expecting a response
            self._resp_queue.put_nowait(cmd_resp_pair)

        compl_obj.set_completed()

    def _process_data_packet(self):
        packet = DataPacket(self._stream_types)
        packet.receive(self._serial)

        for cb in self._stream_cbs:
            cb(packet)

    def _process_in_stream_resp(self):
        peek = self._serial.peek(len(FULL_STATUS_RESPONSE))

        if peek == FULL_STATUS_RESPONSE:
            # The packet is a status response, which we need to handle separately
            self._process_status_response()
        else:
            # We don't know exactly what it is, but it must have been triggered by a command, so we simply
            # process it as part of the reqular queue handling
            self._process_resp_from_queue()

    def _process_status_response(self):
        cmd_resp_pair = self._resp_queue.peek()

        if cmd_resp_pair is not None and isinstance(cmd_resp_pair[0], GetStatusCommand):
            # We have received a Status Response and have are expecting a response from a command
            # ---> Handle it like a regular command
            self._process_resp_from_queue()
        else:
            # We have received a Status Response but are not expecting one
            # ---> Handle it as a pushed Status Update
            self._process_status_update()

    def _process_status_update(self):
        # Called if the status response was not triggered by a command but sent by the Shimmer as the result of
        # an event
        status_cmd = GetStatusCommand()
        r = status_cmd.receive(self._serial)

        for cb in self._status_cbs:
            cb(r)

    def _process_resp_from_queue(self):
        cmd, return_obj = self._resp_queue.get_nowait()

        resp_code = cmd.get_response_code()
        peek = self._serial.peek(len(resp_code))

        if peek != resp_code:
            raise ValueError(f'Expecting response code {fmt_hex(resp_code)} but found {fmt_hex(peek)}')

        result = cmd.receive(self._serial)
        return_obj.set_result(result)

    def process_single_input_event(self) -> None:
        """Process and read a single input event

        An input event can be a single acknowledgment or a request response. The function does not return anything.
        All data is provided via the completion objects returned when queueing the command.

        """
        peek = self._serial.peek_packed('B')

        if peek == ACK_COMMAND_PROCESSED:
            self._process_ack()
        elif peek == DATA_PACKET:
            self._process_data_packet()
        elif peek == INSTREAM_CMD_RESPONSE:
            self._process_in_stream_resp()
        else:
            self._process_resp_from_queue()

    def queue_command(self, cmd: ShimmerCommand) -> Tuple[RequestCompletion, RequestResponse]:
        """Queue a command request for processing

        :param cmd: The command to send to the Shimmer device
        :return: A completion instance and a response instance. The completion instance is always returned and becomes
            true when the command has been processed by the Shimmer. The response object is only returned if the command
            features a response. It holds the response data once the response has been returned by the Shimmer.
        """
        resp_obj = None
        cmd_resp_pair = (None, None)
        compl_obj = RequestCompletion()

        if cmd.has_response():
            resp_obj = RequestResponse()
            cmd_resp_pair = (cmd, resp_obj)

        self._ack_queue.put_nowait((compl_obj, cmd_resp_pair))
        cmd.send(self._serial)

        return compl_obj, resp_obj

    def clear_queues(self) -> None:
        """Clear the internal queues and release any locks held by other threads

        """
        try:
            while True:
                compl, (cmd, resp) = self._ack_queue.get_nowait()
                compl.set_completed()

                if resp is not None:
                    resp.set_result(None)
        except Empty:
            pass

        try:
            while True:
                _, resp = self._resp_queue.get_nowait()
                resp.set_result(None)
        except Empty:
            pass


class ShimmerBluetooth:

    def __init__(self, serial: Serial, disable_status_ack: bool = True):
        """API for communicating with the Shimmer via Bluetooth

        This class implements support for talking to the Shimmer LogAndStream firmware via Bluetooth.
        Each command is encapsulated as a method that can be called to invoke the corresponding command.
        All commands are executed synchronously. This means that the method call will block until the
        Shimmer has processed the request and responded.

        :param serial: The serial channel that encapsulates the rfcomm Bluetooth connection to the Shimmer
        :param disable_status_ack: Starting with LogAndStream firmware version 0.15.4, the vanilla firmware
            supports disabling the acknowledgment byte before status messages. This removes the need for
            running a custom firmware version on the Shimmer. If this flag is set to True, the API will
            query the firmware version of the Shimmer and automatically send a command to disable the status
            acknowledgment byte at startup. You can set it to True if you don't want this or if it causes
            trouble with your firmware version.
        """
        self._serial = BluetoothSerial(serial)
        self._bluetooth = BluetoothRequestHandler(self._serial)

        self._thread = Thread(target=self._run_readloop, daemon=True)

        self._initialized = False
        self._disable_ack = disable_status_ack

        self._fw_version: Optional[FirmwareVersion] = None
        self._fw_caps: Optional[FirmwareCapabilities] = None

    @property
    def initialized(self) -> bool:
        """Specifies if the connection was initialized

        This property helps to determine if the capabilities property will return a valid value.

        :return: True if initialize() was called, otherwise False
        """
        return self._initialized

    @property
    def capabilities(self) -> FirmwareCapabilities:
        """Return the capabilities of the device firmware

        This property shall only be accessed after invoking initialize().

        :return: A FirmwareCapabilities instance representing the version and capabilities of the firmware
        """
        return self._fw_caps

    def __enter__(self):
        self.initialize()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.shutdown()

    def _set_fw_capabilities(self) -> None:
        fw_type, fw_ver = self.get_firmware_version()
        self._fw_caps = FirmwareCapabilities(fw_type, fw_ver)

    def initialize(self) -> None:
        """Initialize the Bluetooth connection

        This method must be invoked before sending commands to the Shimmer. It queries the Shimmer version,
        optionally disables the status acknowledgment and starts the read loop.
        """
        self._thread.start()

        self._set_fw_capabilities()

        if self.capabilities.supports_ack_disable and self._disable_ack:
            self.set_status_ack(enabled=False)

        self._initialized = True

    def shutdown(self) -> None:
        """Shutdown the read loop

        Shutdown the read loop by stopping the read thread
        """
        self._serial.cancel_read()
        self._thread.join()
        self._serial.close()
        self._bluetooth.clear_queues()

    def _run_readloop(self):
        try:
            while True:
                self._bluetooth.process_single_input_event()

        except ReadAbort:
            print('Read loop exciting after cancel request')

    def _process_and_wait(self, cmd):
        compl_obj, return_obj = self._bluetooth.queue_command(cmd)
        compl_obj.wait()

        if return_obj is not None:
            return return_obj.wait()
        return None

    def add_stream_callback(self, cb: Callable[[DataPacket], None]) -> None:
        """Add a stream callback which is called when a new data packet arrives

        :param cb: a function with a single argument
        """
        self._bluetooth.add_stream_callback(cb)

    def remove_stream_callback(self, cb: Callable[[DataPacket], None]) -> None:
        """Remove the callback from the list of active callbacks

        :param cb: The callback function to remove
        """
        self._bluetooth.remove_stream_callback(cb)

    def add_status_callback(self, cb: Callable[[List[bool]], None]) -> None:
        """Add a status callback which is called when a new status update from the Shimmer arrives

        :param cb: a function with a single argument. The argument is the same value is the return value of the
            :meth:`pyshimmer.bluetooth.bt_api.ShimmerBluetooth.get_status` method.
        """
        self._bluetooth.add_status_callback(cb)

    def remove_status_callback(self, cb: Callable[[List[bool]], None]) -> None:
        """Remove the callback from the list of active callbacks

        :param cb: The callback function to remove
        """
        self._bluetooth.remove_status_callback(cb)

    def get_sampling_rate(self) -> float:
        """Retrieve the sampling rate of the device

        :return: The sampling rate as floating point value in samples per second
        """
        return self._process_and_wait(GetSamplingRateCommand())

    def set_sampling_rate(self, sr: float) -> None:
        """Set the active sampling rate for the device

        :param sr: The sampling rate in Hertz
        """
        self._process_and_wait(SetSamplingRateCommand(sr))

    def get_battery_state(self, in_percent: bool) -> float:
        """Retrieve the battery state of the device

        :param in_percent: True: calculate battery state in percent; False: calculate battery state in Volt
        :return: The battery state in percent / Volt
        """
        return self._process_and_wait(GetBatteryCommand(in_percent))

    def get_config_time(self) -> int:
        """Get the config time from the device as configured in the configuration file

        :return: The config time as integer
        """
        return self._process_and_wait(GetConfigTimeCommand())

    def set_config_time(self, time: int) -> None:
        """Set the config time of the device

        :arg time: The configuration time that will be set in the configuration of the Shimmer
        """
        self._process_and_wait(SetConfigTimeCommand(time))

    def set_sensors(self, sensors: Iterable[ESensorGroup]) -> None:
        """Set the active sensors for sampling

        This command will activate the specified list of sensors and deactivate all other sensors.

        :param sensors: A list of sensors to activate
        """
        self._process_and_wait(SetSensorsCommand(sensors))

    def get_rtc(self) -> float:
        """Retrieve the current value of the onboard real-time clock

        :return: The current time of the device in seconds as UNIX timestamp
        """
        return self._process_and_wait(GetRealTimeClockCommand())

    def set_rtc(self, time_sec: float) -> None:
        """Set the value of the onboard real-time clock

        Should be set as a UTC UNIX timestamp such that the resulting recordings have universal timestamps

        :param time_sec: The UNIX timestamp in seconds
        """
        self._process_and_wait(SetRealTimeClockCommand(time_sec))

    def get_status(self) -> List[bool]:
        """Get the status of the device

        :return: A list of 8 bools which signal:
            dev_docked, dev_sensing, rtc_set, dev_logging, dev_streaming, sd_card_present, sd_error, status_red_led
        """
        return self._process_and_wait(GetStatusCommand())

    def get_firmware_version(self) -> Tuple[EFirmwareType, FirmwareVersion]:
        """Get the version of the running firmware

        :return: The firmware type as enum, i.e. SDLog or LogAndStream
            and the numeric firmware version
        """
        fw_type, major, minor, rel = self._process_and_wait(GetFirmwareVersionCommand())
        fw_version = FirmwareVersion(major, minor, rel)

        return fw_type, fw_version

    def get_exg_register(self, chip_id: int) -> ExGRegister:
        """Get the current configuration of one of the two ExG registers of the device

        Note that this command only returns meaningful results if the device features ECG chips

        :param chip_id: The ID of the chip, one of [0, 1]
        :return: An ExGRegister object that presents the register contents in an easily processable manner
        """
        return self._process_and_wait(GetEXGRegsCommand(chip_id))

    def get_all_calibration(self) -> AllCalibration:
        """Gets all calibration data from sensor
        :return: An AllCalibration object that presents the calibration contents in an easily processable manner
        """
        return self._process_and_wait(GetAllCalibrationCommand())

    def set_exg_register(self, chip_id: int, offset: int, data: bytes) -> None:
        """Configure part of the memory of the ExG registers

        :param chip_id: The ID of the chip, one of [0, 1]
        :param offset: The offset at which to write the data bytes
        :param data: The data bytes to write
        """
        self._process_and_wait(SetEXGRegsCommand(chip_id, offset, data))

    def get_device_name(self) -> str:
        """Retrieve the device name

        :return: The device name as string
        """
        return self._process_and_wait(GetDeviceNameCommand())

    def set_device_name(self, dev_name: str) -> None:
        """Set the device name

        :param dev_name: The device name to set
        """
        self._process_and_wait(SetDeviceNameCommand(dev_name))

    def get_experiment_id(self) -> str:
        """Retrieve the experiment id as string

        :return: The experiment ID as string
        """
        return self._process_and_wait(GetExperimentIDCommand())

    def set_experiment_id(self, exp_id: str) -> None:
        """Set the experiment ID for the device

        :param exp_id: The id to set for the device
        """
        self._process_and_wait(SetExperimentIDCommand(exp_id))

    def get_inquiry(self) -> Tuple[float, int, List[EChannelType]]:
        """Perform inquiry command

        :return: A tuple of 3 values:
            - The sampling rate as float
            - The buf size of the device
            - The active data channels of the device as list, does not include the TIMESTAMP channel
        """
        return self._process_and_wait(InquiryCommand())

    def get_data_types(self):
        """Get the active data channels of the device

        These data channels will be contained in a DataPacket when streaming

        :return: A list of data channels, always containing the TIMESTAMP channel
        """
        _, _, ctypes = self.get_inquiry()
        # The Timestamp is always present in data packets
        ctypes = [EChannelType.TIMESTAMP] + ctypes

        return ctypes

    def start_streaming(self) -> None:
        """Start streaming data

        """
        ctypes = self.get_data_types()

        stream_types = [(t, ChDataTypeAssignment[t]) for t in ctypes]
        self._bluetooth.set_stream_types(stream_types)

        self._process_and_wait(StartStreamingCommand())

    def stop_streaming(self) -> None:
        """Stop streaming data

        Note that the interface will possibly return more data packets that have already been received and are in the
        input buffer.

        """
        self._process_and_wait(StopStreamingCommand())

    def start_logging(self) -> None:
        """Start logging data to the SD card of the device

        """
        self._process_and_wait(StartLoggingCommand())

    def stop_logging(self) -> None:
        """Stop logging data to the SD card of the device

        """
        self._process_and_wait(StopLoggingCommand())

    def send_ping(self) -> None:
        """Send a ping command to the device

        The command can be used to test the connection. It does not return anything.
        """
        self._process_and_wait(DummyCommand())

    def set_status_ack(self, enabled: bool) -> None:
        """Send a command to enable or disable the status acknowledgment

        This command should normally not be called directly. If enabled in the constructor, the command
        will automatically be sent to the Shimmer if the firmware supports it. It can be used to make
        vanilla firmware versions compatible with the state machine of the Python API.

        :param enabled: If set to True, enable status acknowledgment byte. This will make the
            firmware incompatible to the Python API. If set to False, disable sending the status ack.
            In this state, the firmware is compatible to the Python API.
        """
        self._process_and_wait(SetStatusAckCommand(enabled))



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from typing import List, Set
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase
from unittest.mock import Mock, PropertyMock

import numpy as np
import pandas as pd

from pyshimmer import EChannelType, ExGRegister
from pyshimmer.dev.base import ticks2sec
from pyshimmer.dev.channels import ESensorGroup, get_enabled_channels
from pyshimmer.dev.exg import get_exg_ch
from pyshimmer.reader.binary_reader import ShimmerBinaryReader
from pyshimmer.reader.shimmer_reader import ShimmerReader, SingleChannelProcessor, PPGProcessor, TriAxCalProcessor
from .reader_test_util import get_bin_vs_consensys_pair_fpath, get_synced_bin_vs_consensys_pair_fpath, get_ecg_sample, \
    get_triaxcal_sample


class ShimmerReaderTest(TestCase):

    def test_reader_timestep_interpolation(self):
        sr = 5
        ts_dev = np.array([0, 5, 10, 15, 21, 25, 29, 35])
        ts = ticks2sec(ts_dev)
        vbatt = np.array([93, 85, 78, 74, 71, 68, 65, 64])
        samples = {
            EChannelType.TIMESTAMP: ts_dev,
            EChannelType.VBATT: vbatt,
        }

        m_br = Mock(spec=ShimmerBinaryReader)
        m_br.read_data.return_value = (samples, [])
        type(m_br).sample_rate = PropertyMock(return_value=sr)
        type(m_br).enabled_sensors = PropertyMock(return_value=[])
        type(m_br).has_sync = PropertyMock(return_value=False)
        type(m_br).has_global_clock = PropertyMock(return_value=False)
        type(m_br).start_timestamp = PropertyMock(return_value=0)

        reader = ShimmerReader(bin_reader=m_br)
        reader.load_file_data()

        ts_aligned = reader.timestamp
        self.assertEqual(len(ts_aligned), len(ts))

        vbatt_aligned = reader[EChannelType.VBATT]
        np.testing.assert_equal(vbatt[ts_aligned == ts], vbatt_aligned[ts_aligned == ts])

    def test_timestamp_unwrapping(self):
        sr = 65
        ts_dev = np.arange(0, 4 * (2 ** 24), sr)
        ts_dev_wrapped = ts_dev % 2 ** 24

        ts = ticks2sec(ts_dev)
        vbatt = np.random.randint(0, 100 + 1, len(ts_dev))

        m_br = Mock(spec=ShimmerBinaryReader)
        type(m_br).has_sync = PropertyMock(return_value=False)
        type(m_br).enabled_sensors = PropertyMock(return_value=[])
        type(m_br).sample_rate = PropertyMock(return_value=sr)
        type(m_br).has_global_clock = PropertyMock(return_value=False)
        type(m_br).start_timestamp = PropertyMock(return_value=0)

        samples = {
            EChannelType.VBATT: vbatt,
            EChannelType.TIMESTAMP: ts_dev_wrapped,
        }
        m_br.read_data.return_value = (samples, [])

        reader = ShimmerReader(bin_reader=m_br)
        reader.load_file_data()

        ts_actual = reader.timestamp
        self.assertEqual(len(ts_actual), len(ts))
        np.testing.assert_almost_equal(ts_actual, ts)
        np.testing.assert_equal(reader[EChannelType.VBATT], vbatt)
        np.testing.assert_equal(reader.timestamp, reader[EChannelType.TIMESTAMP])

    def test_timestamp_synchronization(self):
        sr = 5
        ts = np.array([0, 5, 10, 15, 20, 25, 30, 35, 40, 45])
        vbatt = np.array([93, 85, 78, 74, 71, 68, 65, 64, 10, 24])
        samples = {
            EChannelType.TIMESTAMP: ts,
            EChannelType.VBATT: vbatt,
        }

        sync_index = np.array([0, len(ts) - 1])
        sync_offset = np.array([1, 0])

        m_br = Mock(spec=ShimmerBinaryReader)
        m_br.read_data.return_value = (samples, [sync_index, sync_offset])
        type(m_br).sample_rate = PropertyMock(return_value=sr)
        type(m_br).enabled_sensors = PropertyMock(return_value=[])
        type(m_br).has_sync = PropertyMock(return_value=True)
        type(m_br).has_global_clock = PropertyMock(return_value=False)
        type(m_br).start_timestamp = PropertyMock(return_value=0)

        reader = ShimmerReader(bin_reader=m_br)
        reader.load_file_data()

        ts_sync_dev = ts - np.linspace(1, 0, len(ts))
        exp_ts = ticks2sec(ts_sync_dev)
        act_ts = reader.timestamp

        self.assertEqual(len(exp_ts), len(act_ts))
        np.testing.assert_almost_equal(act_ts, exp_ts)
        np.testing.assert_equal(vbatt, reader[EChannelType.VBATT])

    # noinspection PyMethodMayBeStatic
    def test_compare_ppg_processiong_to_consensys(self):
        raw_file, csv_file = get_bin_vs_consensys_pair_fpath()

        exp_sr = 504.12
        exp_channels = [EChannelType.ACCEL_LN_X, EChannelType.ACCEL_LN_Y, EChannelType.ACCEL_LN_Z, EChannelType.VBATT,
                        EChannelType.INTERNAL_ADC_13]

        with open(raw_file, 'rb') as f:
            reader = ShimmerReader(f)
            reader.load_file_data()

        self.assertEqual(exp_channels, reader.channels)
        self.assertAlmostEqual(exp_sr, reader.sample_rate, 2)

        r = np.loadtxt(csv_file, delimiter='\t', skiprows=3, usecols=(0, 1))
        expected_ts = r[:, 0]
        expected_ppg = r[:, 1]

        actual_ts = reader.timestamp * 1000  # needs to be in ms
        actual_ppg = reader[EChannelType.INTERNAL_ADC_13] * 1000.0  # needs to be in mV

        np.testing.assert_almost_equal(actual_ts.flatten(), expected_ts.flatten())
        np.testing.assert_almost_equal(actual_ppg, expected_ppg)

    def test_compare_sync_processing_to_consensys(self):
        bin_path, csv_path = get_synced_bin_vs_consensys_pair_fpath()

        exp_sr = 512.0
        exp_channels = [EChannelType.INTERNAL_ADC_13]

        with open(bin_path, 'rb') as f:
            reader = ShimmerReader(f, sync=True)
            reader.load_file_data()

        csv_data = np.loadtxt(csv_path, delimiter='\t', skiprows=3, usecols=(0, 1))
        expected_ts = csv_data[:, 0]
        expected_ppg = csv_data[:, 1]

        actual_ts = reader.timestamp * 1000
        actual_ppg = reader[EChannelType.INTERNAL_ADC_13] * 1000.0  # needs to be in mV

        self.assertEqual(exp_channels, reader.channels)
        self.assertEqual(exp_sr, reader.sample_rate)
        np.testing.assert_almost_equal(actual_ts.flatten(), expected_ts.flatten())
        np.testing.assert_almost_equal(actual_ppg.flatten(), expected_ppg.flatten())

    def test_reader_exg_register(self):
        exp_reg1_content = bytes(range(10))
        exp_reg1 = ExGRegister(exp_reg1_content)
        exp_reg2_content = bytes(range(10, 0, -1))
        exp_reg2 = ExGRegister(exp_reg2_content)
        exp_regs = [exp_reg1, exp_reg2]

        m_br = Mock(spec=ShimmerBinaryReader)
        m_br.get_exg_reg.side_effect = lambda x: exp_regs[x]
        type(m_br).exg_reg1 = PropertyMock(return_value=exp_reg1)
        type(m_br).exg_reg2 = PropertyMock(return_value=exp_reg2)
        reader = ShimmerReader(bin_reader=m_br)

        for i in range(2):
            self.assertEqual(reader.get_exg_reg(i), exp_regs[i])

        actual_reg1 = reader.exg_reg1
        actual_reg2 = reader.exg_reg2
        self.assertEqual(actual_reg1, exp_reg1)
        self.assertEqual(actual_reg2, exp_reg2)

    # noinspection PyMethodMayBeStatic
    def test_post_process_exg_signal(self):
        exg_reg1 = ExGRegister(b'\x03\x80\x10\x40\x40\x00\x00\x00\x02\x01')
        exg1_gain = 4
        exg_reg2 = ExGRegister(b'\x03\x80\x10\x20\x20\x00\x00\x00\x02\x01')
        exg2_gain = 2

        chip_gain = {
            0: exg1_gain,
            1: exg2_gain,
        }

        samples = {
            EChannelType.EXG_ADS1292R_1_CH1_24BIT: np.random.randn(1000),
            EChannelType.EXG_ADS1292R_2_CH2_24BIT: np.random.randn(1000),
            EChannelType.EXG_ADS1292R_1_CH1_16BIT: np.random.randn(1000),
            EChannelType.EXG_ADS1292R_2_CH2_16BIT: np.random.randn(1000),
        }

        samples_w_ts = {**samples, EChannelType.TIMESTAMP: np.arange(1000)}

        m_br = Mock(spec=ShimmerBinaryReader)
        m_br.get_exg_reg.side_effect = lambda x: exg_reg1 if x == 0 else exg_reg2
        m_br.read_data.side_effect = lambda: (dict(samples_w_ts), ((), ()))
        type(m_br).sample_rate = PropertyMock(return_value=1)
        type(m_br).enabled_sensors = PropertyMock(return_value=[])
        type(m_br).has_sync = PropertyMock(return_value=False)
        type(m_br).has_global_clock = PropertyMock(return_value=False)
        type(m_br).start_timestamp = PropertyMock(return_value=0)
        type(m_br).exg_reg1 = PropertyMock(return_value=exg_reg1)
        type(m_br).exg_reg2 = PropertyMock(return_value=exg_reg2)

        reader = ShimmerReader(bin_reader=m_br, post_process=False)
        reader.load_file_data()
        for ch_type in samples:
            np.testing.assert_equal(samples[ch_type], reader[ch_type])

        reader = ShimmerReader(bin_reader=m_br, post_process=True)
        reader.load_file_data()

        for ch in samples:
            bit = 16 if '16' in ch.name else 24
            gain = chip_gain[get_exg_ch(ch)[0]]
            expected = (samples[ch] - 0) * 2.420 / (2 ** (bit - 1) - 1) / gain
            actual = reader[ch]
            np.testing.assert_almost_equal(actual, expected)

    # noinspection PyMethodMayBeStatic
    def test_compare_exg_processing_to_consensys(self):
        bin_path, uncal_path, cal_path = get_ecg_sample()

        def verify(bin_file_path, expected, post_process):
            with open(bin_file_path, 'rb') as f:
                reader = ShimmerReader(f, post_process=post_process, sync=False)
                reader.load_file_data()

            actual = reader[EChannelType.EXG_ADS1292R_1_CH1_24BIT]
            np.testing.assert_almost_equal(actual, expected[1])

            actual = reader[EChannelType.EXG_ADS1292R_1_CH2_24BIT]
            np.testing.assert_almost_equal(actual, expected[2])

        expected_uncal = np.loadtxt(uncal_path, delimiter='\t', skiprows=3, usecols=(0, 1, 2)).T
        expected_cal = np.loadtxt(cal_path, delimiter='\t', skiprows=3, usecols=(0, 1, 2)).T / 1000.0  # Volt

        verify(bin_path, expected_uncal, post_process=False)
        verify(bin_path, expected_cal, post_process=True)

    # noinspection PyMethodMayBeStatic
    def test_compare_triaxcal_to_consensys(self):
        bin_path, uncal_path, cal_path = get_triaxcal_sample()

        consensys_csv = pd.read_csv(cal_path, sep=",", skiprows=(0, 2), usecols=list(range(14)))
        col_mapping = {
            EChannelType.ACCEL_LN_X: "Shimmer_952D_Accel_LN_X_CAL",
            EChannelType.ACCEL_LN_Y: "Shimmer_952D_Accel_LN_Y_CAL",
            EChannelType.ACCEL_LN_Z: "Shimmer_952D_Accel_LN_Z_CAL",
            EChannelType.ACCEL_LSM303DLHC_X: "Shimmer_952D_Accel_WR_X_CAL",
            EChannelType.ACCEL_LSM303DLHC_Y: "Shimmer_952D_Accel_WR_Y_CAL",
            EChannelType.ACCEL_LSM303DLHC_Z: "Shimmer_952D_Accel_WR_Z_CAL",
            EChannelType.GYRO_MPU9150_X: "Shimmer_952D_Gyro_X_CAL",
            EChannelType.GYRO_MPU9150_Y: "Shimmer_952D_Gyro_Y_CAL",
            EChannelType.GYRO_MPU9150_Z: "Shimmer_952D_Gyro_Z_CAL",
            EChannelType.MAG_LSM303DLHC_X: "Shimmer_952D_Mag_X_CAL",
            EChannelType.MAG_LSM303DLHC_Y: "Shimmer_952D_Mag_Y_CAL",
            EChannelType.MAG_LSM303DLHC_Z: "Shimmer_952D_Mag_Z_CAL",
        }

        with open(bin_path, 'rb') as f:
            reader = ShimmerReader(f)
            reader.load_file_data()

            for rdr_col, csv_col in col_mapping.items():
                rdr_channel = reader[rdr_col]
                csv_channel = consensys_csv[csv_col]
                np.testing.assert_almost_equal(rdr_channel, csv_channel.to_numpy())


class SignalPostProcessorTest(TestCase):

    # noinspection PyTypeChecker
    def test_single_channel_processor(self):
        class TestProcessor(SingleChannelProcessor):

            def __init__(self, channels: List[EChannelType] = None):
                super().__init__(channels)

                self._seen = []

            def process_channel(self, ch_type: EChannelType, y: np.ndarray, reader: ShimmerBinaryReader) -> np.ndarray:
                self._seen.append(ch_type)
                return y

            @property
            def seen(self) -> Set[EChannelType]:
                return set(self._seen)

        ch_data = {
            EChannelType.TIMESTAMP: np.random.randn(10),
            EChannelType.VBATT: np.random.randn(10),
            EChannelType.INTERNAL_ADC_13: np.random.randn(10),
            EChannelType.ACCEL_LN_X: np.random.randn(10),
        }
        ch_types = set(ch_data.keys())

        proc = TestProcessor()
        proc.process(ch_data, None)
        self.assertEqual(proc.seen, ch_types)

        proc = TestProcessor([EChannelType.VBATT])
        proc.process(ch_data, None)
        self.assertEqual(proc.seen, {EChannelType.VBATT})

        proc = TestProcessor([EChannelType.VBATT, EChannelType.ACCEL_LN_Y])
        proc.process(ch_data, None)
        self.assertEqual(proc.seen, {EChannelType.VBATT})

        proc = TestProcessor([EChannelType.VBATT, EChannelType.ACCEL_LN_X])
        proc.process(ch_data, None)
        self.assertEqual(proc.seen, {EChannelType.VBATT, EChannelType.ACCEL_LN_X})

    # noinspection PyMethodMayBeStatic
    def test_ppg_processor(self):
        ppg_data = np.random.randn(10)
        ch_data = {
            EChannelType.TIMESTAMP: np.random.randn(10),
            EChannelType.VBATT: np.random.randn(10),
            EChannelType.INTERNAL_ADC_13: ppg_data,
            EChannelType.ACCEL_LN_X: np.random.randn(10),
        }

        proc = PPGProcessor()
        # noinspection PyTypeChecker
        output = proc.process(ch_data, None)

        for ch, y in output.items():
            if ch != EChannelType.INTERNAL_ADC_13:
                np.testing.assert_equal(y, ch_data[ch])
            else:
                np.testing.assert_equal(y, ppg_data / 1000.0)

    # noinspection PyMethodMayBeStatic
    def test_triaxcal_processor(self):
        o, g, a = np.array([1, 2, 3]), np.diag([4, 5, 6]), np.diag([7, 8, 9])
        params = {ESensorGroup.ACCEL_LN: (o, g, a)}

        ch_types = get_enabled_channels(list(params.keys()))

        data_arr = np.random.randn(3, 100)
        data_dict = {c: data_arr[i] for i, c in enumerate(ch_types)}

        mock_reader = Mock(spec=ShimmerBinaryReader)
        mock_reader.get_triaxcal_params.side_effect = lambda x: params[x]
        type(mock_reader).enabled_sensors = PropertyMock(return_value=list(params.keys()))

        proc = TriAxCalProcessor()
        actual_dict = proc.process(data_dict, mock_reader)
        actual_arr = np.stack([actual_dict[c] for c in ch_types])

        k = np.matmul(np.linalg.inv(a), np.linalg.inv(g))
        exp_arr = np.matmul(k, data_arr - o[..., None])

        np.testing.assert_almost_equal(actual_arr, exp_arr)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase

import numpy as np

from pyshimmer.reader.shimmer_reader import unwrap_device_timestamps, fit_linear_1d


# noinspection PyMethodMayBeStatic
class UtilTest(TestCase):

    def test_unwrap_device_timestamps(self):
        ts_wrapped = np.array([0, 1, 2, 2 ** 24 - 1, 0, 2 ** 24])
        expected = np.array([0, 1, 2, 2 ** 24 - 1, 2 ** 24, 2 * 2 ** 24])
        actual = unwrap_device_timestamps(ts_wrapped)
        np.testing.assert_equal(actual, expected)

        ts_wrapped = np.array([0, 10, 20, 30, 5, 15, 25, 35])
        expected = np.array([0, 10, 20, 30, 5 + 2 ** 24, 15 + 2 ** 24, 25 + 2 ** 24, 35 + 2 ** 24])
        actual = unwrap_device_timestamps(ts_wrapped)
        np.testing.assert_equal(actual, expected)

    def test_fit_linear_1d(self):
        x = np.array([0, 1])
        y = np.array([0, 10])
        xi = np.array([0, 0.25, 0.5, 0.75, 1])

        yi_expected = np.array([0, 2.5, 5, 7.5, 10])
        yi_actual = fit_linear_1d(x, y, xi)
        np.testing.assert_almost_equal(yi_actual, yi_expected)

        xi = 0.1
        yi_expected = 1.0
        yi_actual = fit_linear_1d(x, y, xi)
        np.testing.assert_almost_equal(yi_actual, yi_expected)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from concurrent.futures import ThreadPoolExecutor, Future
from typing import Optional, BinaryIO, List, Callable
from unittest import TestCase

from pyshimmer.bluetooth.bt_api import BluetoothRequestHandler, ShimmerBluetooth
from pyshimmer.bluetooth.bt_commands import GetDeviceNameCommand, SetDeviceNameCommand, DataPacket, GetStatusCommand, \
    GetStringCommand, ResponseCommand
from pyshimmer.bluetooth.bt_serial import BluetoothSerial
from pyshimmer.dev.channels import ChDataTypeAssignment, EChannelType
from pyshimmer.dev.fw_version import FirmwareVersion, EFirmwareType
from pyshimmer.test_util import PTYSerialMockCreator


class BluetoothRequestHandlerTest(TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._mock_creator: Optional[PTYSerialMockCreator] = None
        self._sot: Optional[BluetoothRequestHandler] = None

        self._master: Optional[BinaryIO] = None

    def setUp(self) -> None:
        self._mock_creator = PTYSerialMockCreator()
        serial, self._master = self._mock_creator.create_mock()

        bt_serial = BluetoothSerial(serial)
        self._sot = BluetoothRequestHandler(bt_serial)

    def read_from_master(self, n: int) -> bytes:
        result = bytes()
        while len(result) < n:
            result += self._master.read(n - len(result))

        return result

    def tearDown(self) -> None:
        self._mock_creator.close()

    def test_add_remove_stream_cb(self):
        def cb(_):
            pass

        self._sot.add_stream_callback(cb)
        self._sot.remove_stream_callback(cb)

    def test_add_remove_status_cb(self):
        def cb(_):
            pass

        self._sot.add_status_callback(cb)
        self._sot.remove_status_callback(cb)

    def test_enque_command(self):
        cmd = GetDeviceNameCommand()
        compl, resp = self._sot.queue_command(cmd)

        self.assertFalse(compl.has_completed())
        self.assertFalse(resp.has_result())

        r = self.read_from_master(1)
        self.assertEqual(r, b'\x7b')

        self._master.write(b'\xff')
        self._sot.process_single_input_event()

        self.assertTrue(compl.has_completed())
        self.assertFalse(resp.has_result())

        self._master.write(b'\x7a\x05\x53\x5f\x50\x50\x47')
        self._sot.process_single_input_event()

        self.assertTrue(resp.has_result())
        self.assertEqual(resp.get_result(), 'S_PPG')

    def test_enqueue_multibyte(self):
        cmd = GetStringCommand(0x10, b'\x0a\x0b')
        compl, resp = self._sot.queue_command(cmd)

        r = self.read_from_master(1)
        self.assertEqual(r, b'\x10')

        self._master.write(b'\xff')
        self._sot.process_single_input_event()

        self._master.write(b'\x0a\x0b\x02ab')
        self._sot.process_single_input_event()

        self.assertTrue(compl.has_completed())
        self.assertTrue(resp.has_result())
        self.assertEqual(resp.get_result(), 'ab')

    def test_enqueue_multiple_commands(self):
        cmd1 = GetDeviceNameCommand()
        cmd2 = GetStatusCommand()

        compl1, resp1 = self._sot.queue_command(cmd1)
        compl2, resp2 = self._sot.queue_command(cmd2)

        r = self.read_from_master(2)
        self.assertEqual(r, b'\x7B\x72')

        self._master.write(b'\xff\x7a\x05\x53\x5f\x50\x50\x47')
        self._master.write(b'\xff\x8a\x71\x21')

        self._sot.process_single_input_event()
        self.assertTrue(compl1.has_completed())

        self._sot.process_single_input_event()
        self.assertTrue(resp1.has_result())
        self.assertEqual(resp1.get_result(), 'S_PPG')

        self._sot.process_single_input_event()
        self.assertTrue(compl2.has_completed())

        self._sot.process_single_input_event()
        self.assertTrue(resp2.has_result())
        self.assertEqual(resp2.get_result(), [True, False, False, False, False, True, False, False])

    def test_queue_command_no_resp(self):
        cmd = SetDeviceNameCommand('S_PPG')
        compl, resp = self._sot.queue_command(cmd)

        self.assertFalse(compl.has_completed())
        self.assertEqual(resp, None)

        r = self.read_from_master(7)
        self.assertEqual(r, b'\x79\x05S_PPG')

        self._master.write(b'\xff')
        self._sot.process_single_input_event()
        self.assertTrue(compl.has_completed())

    def test_queue_unknown_instream(self):
        class InStreamCommand(ResponseCommand):

            def __init__(self):
                super().__init__(b'\x8a\x42')

            def send(self, ser: BluetoothSerial) -> None:
                ser.write(b'\x42')

            def receive(self, ser: BluetoothSerial) -> any:
                return ser.read_response(b'\x8a\x42')

        compl, resp = self._sot.queue_command(InStreamCommand())

        r = self.read_from_master(1)
        self.assertEqual(r, b'\x42')

        self._master.write(b'\xff\x8a\x42')
        self._sot.process_single_input_event()
        self.assertTrue(compl.has_completed())

        self._sot.process_single_input_event()
        self.assertTrue(resp.has_result())

    def test_get_status_command(self):
        cmd = GetStatusCommand()
        compl, resp = self._sot.queue_command(cmd)

        self.assertFalse(compl.has_completed())
        self.assertFalse(resp.has_result())

        r = self.read_from_master(1)
        self.assertEqual(r, b'\x72')

        self._master.write(b'\xff\x8a\x71\x21')
        self._sot.process_single_input_event()
        self.assertTrue(compl.has_completed())
        self.assertFalse(resp.has_result())

        self._sot.process_single_input_event()
        self.assertTrue(resp.has_result())
        self.assertEqual(resp.get_result(), [True, False, False, False, False, True, False, False])

    def test_incorrect_resp_code_fail(self):
        cmd = GetDeviceNameCommand()
        _ = self._sot.queue_command(cmd)

        self._master.write(b'\xff\xfe')
        self._sot.process_single_input_event()
        self.assertRaises(ValueError, self._sot.process_single_input_event)

    def test_data_packet(self):
        results: List[DataPacket] = []

        data_pkt_1 = b'\x00\xde\xd0\xb2\x26\x07'
        data_pkt_2 = b'\x00\x1e\xd1\xb2\xfc\x06'

        ch_types = [EChannelType.TIMESTAMP, EChannelType.INTERNAL_ADC_13]
        self._sot.set_stream_types([(c, ChDataTypeAssignment[c]) for c in ch_types])
        self._sot.add_stream_callback(results.append)

        self._master.write(data_pkt_1)
        self._master.write(data_pkt_2)

        self._sot.process_single_input_event()
        self.assertEqual(len(results), 1)
        pkt = results[0]

        self.assertEqual(pkt.channels, ch_types)
        self.assertEqual(pkt[EChannelType.TIMESTAMP], 0xb2d0de)
        self.assertEqual(pkt[EChannelType.INTERNAL_ADC_13], 0x0726)

        self._sot.process_single_input_event()
        self.assertEqual(len(results), 2)
        pkt = results[1]

        self.assertEqual(pkt.channels, ch_types)
        self.assertEqual(pkt[EChannelType.TIMESTAMP], 0xb2d11e)
        self.assertEqual(pkt[EChannelType.INTERNAL_ADC_13], 0x06fc)

    def test_get_status_response(self):
        status_resp: List[List[bool]] = []

        stat_pkt_1 = b'\x8a\x71\x20'
        stat_pkt_2 = b'\x8a\x71\x21'
        self._sot.add_status_callback(status_resp.append)

        self._master.write(stat_pkt_1)
        self._master.write(stat_pkt_2)

        self._sot.process_single_input_event()
        self.assertEqual(len(status_resp), 1)
        self.assertEqual(status_resp[0], [False, False, False, False, False, True, False, False])

        self._sot.process_single_input_event()
        self.assertEqual(len(status_resp), 2)
        self.assertEqual(status_resp[1], [True, False, False, False, False, True, False, False])

    def test_get_status_response_update_mixed(self):
        stat_pkt_1 = b'\x8a\x71\x20'
        stat_pkt_2 = b'\x8a\x71\x21'

        status_resp: List[List[bool]] = []
        self._sot.add_status_callback(status_resp.append)

        compl, resp = self._sot.queue_command(GetStatusCommand())
        r = self.read_from_master(1)
        self.assertEqual(r, b'\x72')

        self._master.write(b'\xff' + stat_pkt_1)
        self._master.write(stat_pkt_2)

        self._sot.process_single_input_event()
        self.assertTrue(compl.has_completed())

        self._sot.process_single_input_event()
        self.assertTrue(resp.has_result())
        self.assertEqual(resp.get_result(), [False, False, False, False, False, True, False, False])

        self._sot.process_single_input_event()
        self.assertEqual(len(status_resp), 1)
        self.assertEqual(status_resp[0], [True, False, False, False, False, True, False, False])

    def test_clear_queues(self):
        compl1, resp1 = self._sot.queue_command(GetDeviceNameCommand())
        compl2, resp2 = self._sot.queue_command(GetDeviceNameCommand())

        self.assertFalse(compl1.has_completed())
        self.assertFalse(resp1.has_result())

        # Ensure that the first command has been passed into the response queue
        self._master.write(b'\xff')
        self._sot.process_single_input_event()

        self.assertTrue(compl1.has_completed())
        self.assertFalse(resp1.has_result())

        self.assertFalse(compl2.has_completed())
        self.assertFalse(resp2.has_result())

        self._sot.clear_queues()

        self.assertTrue(compl1.has_completed())
        self.assertTrue(resp1.has_result())
        self.assertEqual(resp1.get_result(), None)

        self.assertTrue(compl2.has_completed())
        self.assertTrue(resp2.has_result())
        self.assertEqual(resp2.get_result(), None)


class ShimmerBluetoothIntegrationTest(TestCase):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._executor = ThreadPoolExecutor(max_workers=1)

        self._mock_creator: Optional[PTYSerialMockCreator] = None
        self._sot: Optional[ShimmerBluetooth] = None

        self._master: Optional[BinaryIO] = None

    def _submit_handler_fn(self, fn: Callable[[BinaryIO, ShimmerBluetooth], any]) -> Future:
        return self._executor.submit(fn, self._master, self._sot)

    def _submit_req_resp_handler(self, req_len: int, resp: bytes) -> Future:
        def master_fn(master: BinaryIO, _) -> bytes:
            req = bytes()
            while len(req) < req_len:
                req += master.read(req_len - len(req))

            master.write(resp)
            return req

        return self._submit_handler_fn(master_fn)

    def do_setup(self, initialize: bool = True, **kwargs) -> None:
        self._mock_creator = PTYSerialMockCreator()
        serial, self._master = self._mock_creator.create_mock()

        self._sot = ShimmerBluetooth(serial, **kwargs)

        if initialize:
            # The Bluetooth API automatically requests the firmware version upon initialization.
            # We must prepare a proper response beforehand.
            future = self._submit_req_resp_handler(req_len=1, resp=b'\xff\x2f\x03\x00\x00\x00\x0b\x00')
            self._sot.initialize()

            # Check that it properly asked for the firmware version
            result = future.result()
            assert result == b'\x2E'

    def tearDown(self) -> None:
        self._sot.shutdown()
        self._mock_creator.close()

    def test_context_manager(self):
        self.do_setup(initialize=False)

        # The Bluetooth API automatically requests the firmware version upon initialization.
        # We must prepare a proper response beforehand.
        req_future = self._submit_req_resp_handler(req_len=1, resp=b'\xff\x2f\x03\x00\x00\x00\x0b\x00')
        with self._sot:
            # We check that the API properly asked for the firmware version
            req_data = req_future.result()
            self.assertEqual(req_data, b'\x2e')

            # It should now be in an initialized state
            self.assertTrue(self._sot.initialized)

    def test_version_and_capabilities(self):
        self.do_setup(initialize=True)

        self.assertTrue(self._sot.initialized)
        self.assertIsNotNone(self._sot.capabilities)
        self.assertEqual(self._sot.capabilities.fw_type, EFirmwareType.LogAndStream)
        self.assertEqual(self._sot.capabilities.version, FirmwareVersion(0, 11, 0))

    def test_get_sampling_rate(self):
        self.do_setup()

        ftr = self._submit_req_resp_handler(1, b'\xff\x04\x40\x00')
        r = self._sot.get_sampling_rate()

        self.assertEqual(ftr.result(), b'\x03')
        self.assertEqual(r, 512.0)

    def test_get_data_types(self):
        self.do_setup()

        ftr = self._submit_req_resp_handler(1, b'\xff\x02\x40\x00\x01\xff\x01\x09\x01\x01\x12')
        r = self._sot.get_data_types()

        self.assertEqual(ftr.result(), b'\x01')
        self.assertEqual(r, [EChannelType.TIMESTAMP, EChannelType.INTERNAL_ADC_13])

    def test_streaming(self):
        self.do_setup()

        pkts = []

        def pkt_handler(new_pkt: DataPacket) -> None:
            pkts.append(new_pkt)

        inquiry_ftr = self._submit_req_resp_handler(1, b'\xff\x02\x40\x00\x01\xff\x01\x09\x01\x01\x12')
        start_streaming_ftr = self._submit_req_resp_handler(1, b'\xff')
        self._submit_req_resp_handler(0, b'\x00\x25\x13\xf4\x4a\x07')
        stop_streaming_ftr = self._submit_req_resp_handler(1, b'\xff')

        self._sot.add_stream_callback(pkt_handler)
        self._sot.start_streaming()

        self.assertEqual(inquiry_ftr.result(), b'\x01')
        self.assertEqual(start_streaming_ftr.result(), b'\x07')

        self._sot.stop_streaming()
        self.assertEqual(stop_streaming_ftr.result(), b'\x20')

        self.assertEqual(len(pkts), 1)
        pkt = pkts[0]

        self.assertEqual(pkt[EChannelType.TIMESTAMP], 15995685)
        self.assertEqual(pkt[EChannelType.INTERNAL_ADC_13], 1866)

    def test_status_update(self):
        self.do_setup()

        pkts = []

        def status_handler(new_pkt: List[bool]) -> None:
            pkts.append(new_pkt)

        self._sot.add_status_callback(status_handler)

        self._submit_req_resp_handler(1, b'\x8a\x71\x20\xff\x7a\x03ABC')
        r = self._sot.get_device_name()
        self.assertEqual(r, 'ABC')

        self.assertEqual(len(pkts), 1)
        pkt = pkts[0]

        self.assertEqual(pkt, [False, False, False, False, False, True, False, False])

    def test_get_firmware_version(self):
        self.do_setup()

        self._submit_req_resp_handler(1, b'\xFF\x2F\x03\x00\x01\x00\x02\x03')
        fwtype, fwver = self._sot.get_firmware_version()

        self.assertEqual(fwtype, EFirmwareType.LogAndStream)
        self.assertEqual(fwver, FirmwareVersion(1, 2, 3))

    def test_status_ack_disable(self):
        self.do_setup(initialize=False)

        # Queue response for version command
        self._submit_req_resp_handler(1, b'\xFF\x2F\x03\x00\x00\x00\x0F\x04')
        # Queue response for disabling the status acknowledgment
        req_future = self._submit_req_resp_handler(2, b'\xFF')

        self._sot.initialize()
        req_data = req_future.result()
        self.assertEqual(req_data, b'\xA3\x00')

    def test_status_ack_not_disable(self):
        self.do_setup(initialize=False, disable_status_ack=False)

        # Queue response for version command
        self._submit_req_resp_handler(1, b'\xFF\x2F\x03\x00\x00\x00\x0F\x04')
        self._sot.initialize()



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Tuple, Union
from unittest import TestCase

from pyshimmer.bluetooth.bt_commands import ShimmerCommand, GetSamplingRateCommand, GetBatteryCommand, \
    GetConfigTimeCommand, SetConfigTimeCommand, GetRealTimeClockCommand, SetRealTimeClockCommand, GetStatusCommand, \
    GetFirmwareVersionCommand, InquiryCommand, StartStreamingCommand, StopStreamingCommand, StartLoggingCommand, \
    StopLoggingCommand, GetEXGRegsCommand, SetEXGRegsCommand, GetExperimentIDCommand, SetExperimentIDCommand, \
    GetDeviceNameCommand, SetDeviceNameCommand, DummyCommand, DataPacket, ResponseCommand, SetStatusAckCommand, \
    SetSensorsCommand, SetSamplingRateCommand, GetAllCalibrationCommand
from pyshimmer.bluetooth.bt_serial import BluetoothSerial
from pyshimmer.dev.channels import ChDataTypeAssignment, EChannelType, ESensorGroup
from pyshimmer.dev.fw_version import EFirmwareType
from pyshimmer.test_util import MockSerial


class BluetoothCommandsTest(TestCase):

    @staticmethod
    def create_mock() -> Tuple[BluetoothSerial, MockSerial]:
        mock = MockSerial()
        # noinspection PyTypeChecker
        serial = BluetoothSerial(mock)
        return serial, mock

    def assert_cmd(self, cmd: ShimmerCommand, req_data: bytes,
                   resp_code: bytes = None, resp_data: bytes = None, exp_result: any = None) -> any:
        serial, mock = self.create_mock()

        cmd.send(serial)
        actual_req_data = mock.test_get_write_data()
        self.assertEqual(actual_req_data, req_data)

        if resp_code is None:
            self.assertFalse(cmd.has_response())
            return None

        self.assertTrue(cmd.has_response())
        self.assertEqual(cmd.get_response_code(), resp_code)

        mock.test_put_read_data(resp_data)

        act_result = cmd.receive(serial)
        if exp_result is not None:
            self.assertEqual(act_result, exp_result)
        return act_result

    def test_response_command_code_conversion(self):
        class TestCommand(ResponseCommand):
            def __init__(self, rcode: Union[int, Tuple[int, ...], bytes]):
                super().__init__(rcode)

            def send(self, ser: BluetoothSerial) -> None:
                pass

        cmd = TestCommand(10)
        self.assertEqual(cmd.get_response_code(), b'\x0a')

        cmd = TestCommand(20)
        self.assertEqual(cmd.get_response_code(), b'\x14')

        cmd = TestCommand((10,))
        self.assertEqual(cmd.get_response_code(), b'\x0a')

        cmd = TestCommand((10, 20))
        self.assertEqual(cmd.get_response_code(), b'\x0a\x14')

        cmd = TestCommand(b'\x10')
        self.assertEqual(cmd.get_response_code(), b'\x10')

        cmd = TestCommand(b'\x10\x20')
        self.assertEqual(cmd.get_response_code(), b'\x10\x20')

    def test_get_sampling_rate_command(self):
        cmd = GetSamplingRateCommand()
        self.assert_cmd(cmd, b'\x03', b'\x04', b'\x04\x40\x00', 512.0)

    def test_set_sampling_rate_command(self):
        cmd = SetSamplingRateCommand(sr=512.0)
        self.assert_cmd(cmd, b'\x05\x40\x00')

    def test_get_battery_state_command(self):
        cmd = GetBatteryCommand(in_percent=True)
        self.assert_cmd(cmd, b'\x95', b'\x8a\x94',
                        b'\x8a\x94\x30\x0b\x80', 100)

        cmd = GetBatteryCommand(in_percent=False)
        self.assert_cmd(cmd, b'\x95', b'\x8a\x94',
                        b'\x8a\x94\x2e\x0b\x80', 4.168246153846154)

    def test_set_sensors_command(self):
        sensors = [
            ESensorGroup.GYRO,
            ESensorGroup.CH_A13,
            ESensorGroup.PRESSURE,
        ]
        cmd = SetSensorsCommand(sensors)
        self.assert_cmd(cmd, b'\x08\x40\x01\x04')

    def test_get_config_time_command(self):
        cmd = GetConfigTimeCommand()
        self.assert_cmd(cmd, b'\x87', b'\x86', b'\x86\x02\x34\x32', 42)

    def test_set_config_time_command(self):
        cmd = SetConfigTimeCommand(43)
        self.assert_cmd(cmd, b'\x85\x02\x34\x33')

    def test_get_rtc(self):
        cmd = GetRealTimeClockCommand()
        r = self.assert_cmd(cmd, b'\x91', b'\x90',
                            b'\x90\x1f\xb1\x93\x09\x00\x00\x00\x00')
        self.assertAlmostEqual(r, 4903.3837585)

    def test_set_rtc(self):
        cmd = SetRealTimeClockCommand(10)
        self.assert_cmd(cmd, b'\x8f\x00\x00\x05\x00\x00\x00\x00\x00')

    def test_get_status_command(self):
        cmd = GetStatusCommand()
        expected_result = [True, False, True, False, False, True, False, False]
        self.assert_cmd(cmd, b'\x72', b'\x8a\x71',
                        b'\x8a\x71\x25', expected_result)

    def test_get_firmware_version_command(self):
        cmd = GetFirmwareVersionCommand()
        fw_type, major, minor, patch = self.assert_cmd(
            cmd, b'\x2e', b'\x2f', b'\x2f\x03\x00\x00\x00\x0b\x00')
        self.assertEqual(fw_type, EFirmwareType.LogAndStream)
        self.assertEqual(major, 0)
        self.assertEqual(minor, 11)
        self.assertEqual(patch, 0)

    def test_inquiry_command(self):
        cmd = InquiryCommand()
        sr, buf_size, ctypes = self.assert_cmd(
            cmd, b'\x01', b'\x02', b'\x02\x40\x00\x01\xff\x01\x09\x01\x01\x12')

        self.assertEqual(sr, 512.0)
        self.assertEqual(buf_size, 1)
        self.assertEqual(ctypes, [EChannelType.INTERNAL_ADC_13])

    def test_start_streaming_command(self):
        cmd = StartStreamingCommand()
        self.assert_cmd(cmd, b'\x07')

    def test_stop_streaming_command(self):
        cmd = StopStreamingCommand()
        self.assert_cmd(cmd, b'\x20')

    def test_start_logging_command(self):
        cmd = StartLoggingCommand()
        self.assert_cmd(cmd, b'\x92')

    def test_stop_logging_command(self):
        cmd = StopLoggingCommand()
        self.assert_cmd(cmd, b'\x93')

    def test_get_exg_register_command(self):
        cmd = GetEXGRegsCommand(1)
        r = self.assert_cmd(cmd, b'\x63\x01\x00\x0a', b'\x62',
                            b'\x62\x0a\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')
        self.assertEqual(r.binary, b'\x00\x80\x10\x00\x00\x00\x00\x00\x02\x01')

    def test_get_exg_reg_fail(self):
        serial, mock = self.create_mock()
        cmd = GetEXGRegsCommand(1)

        mock.test_put_read_data(b'\x62\x04\x01\x02\x03\x04')
        self.assertRaises(ValueError, cmd.receive, serial)

    def test_get_allcalibration_command(self):
        cmd = GetAllCalibrationCommand()
        r = self.assert_cmd(cmd, b'\x2c', b'\x2d', b'\x2d\x08\xcd\x08\xcd\x08\xcd\x00\x5c\x00\x5c\x00\x5c\x00\x9c\x00\x9c\x00\x00\x00\x00\x9c\x00\x00\x00\x00\x00\x00\x19\x96\x19\x96\x19\x96\x00\x9c\x00\x9c\x00\x00\x00\x00\x9c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x87\x06\x87\x06\x87\x00\x9c\x00\x64\x00\x00\x00\x00\x9c')
        self.assertEqual(r.binary, b'\x08\xcd\x08\xcd\x08\xcd\x00\x5c\x00\x5c\x00\x5c\x00\x9c\x00\x9c\x00\x00\x00\x00\x9c\x00\x00\x00\x00\x00\x00\x19\x96\x19\x96\x19\x96\x00\x9c\x00\x9c\x00\x00\x00\x00\x9c\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x06\x87\x06\x87\x06\x87\x00\x9c\x00\x64\x00\x00\x00\x00\x9c')

    def test_set_exg_register_command(self):
        cmd = SetEXGRegsCommand(1, 0x02, b'\x10\x00')
        self.assert_cmd(cmd, b'\x61\x01\x02\x02\x10\x00')

    def test_get_experiment_id_command(self):
        cmd = GetExperimentIDCommand()
        self.assert_cmd(cmd, b'\x7e', b'\x7d', b'\x7d\x06a_test', 'a_test')

    def test_set_experiment_id_command(self):
        cmd = SetExperimentIDCommand('A_Test')
        self.assert_cmd(cmd, b'\x7c\x06A_Test')

    def test_get_device_name_command(self):
        cmd = GetDeviceNameCommand()
        self.assert_cmd(cmd, b'\x7b', b'\x7a', b'\x7a\x05S_PPG', 'S_PPG')

    def test_set_device_name_command(self):
        cmd = SetDeviceNameCommand('S_PPG')
        self.assert_cmd(cmd, b'\x79\x05S_PPG')

    def test_set_status_ack_command(self):
        cmd = SetStatusAckCommand(enabled=True)
        self.assert_cmd(cmd, b'\xA3\x01')

        cmd = SetStatusAckCommand(enabled=False)
        self.assert_cmd(cmd, b'\xA3\x00')

    def test_dummy_command(self):
        cmd = DummyCommand()
        self.assert_cmd(cmd, b'\x96')

    def test_data_packet(self):
        serial, mock = self.create_mock()

        channels = [EChannelType.TIMESTAMP, EChannelType.INTERNAL_ADC_13]
        data_types = [ChDataTypeAssignment[c] for c in channels]
        ch_and_types = list(zip(channels, data_types))

        pkt = DataPacket(ch_and_types)
        self.assertEqual(pkt.channels, channels)
        self.assertEqual(pkt.channel_types, data_types)

        mock.test_put_read_data(b'\x00\xde\xd0\xb2\x26\x07')
        pkt.receive(serial)

        self.assertEqual(pkt[EChannelType.TIMESTAMP], 0xb2d0de)
        self.assertEqual(pkt[EChannelType.INTERNAL_ADC_13], 0x0726)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from typing import Union, Tuple

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from serial import Serial

from pyshimmer.bluetooth.bt_const import ACK_COMMAND_PROCESSED
from pyshimmer.serial_base import SerialBase
from pyshimmer.util import fmt_hex, resp_code_to_bytes


class BluetoothSerial(SerialBase):
    """Another wrapper layer around the Serial interface

    :arg serial: The :class:`serial.Serial` instance to wrap
    """

    def __init__(self, serial: Serial):
        super().__init__(serial)

    def read_varlen(self) -> bytes:
        """Read the number of bytes specified by the first byte read

        The function reads a single byte that it interprets as a length field with value L and then reads the
        specified number of L bytes from the stream.

        :return: A variable number of bytes
        """
        arg_len = self.read_byte()
        arg = self.read(arg_len)
        return arg

    def write_varlen(self, arg: bytes) -> None:
        """Write a variable number of bytes by prepending the data with their length

        The function first writes the length of the data as a single byte, then writes the data itself.

        :param arg: The data to write
        :raises ValueError: If the number of bytes exceeds the maximum length of 256
        """
        arg_len = len(arg)
        if arg_len > 255:
            raise ValueError(f'Variable-length argument is too long: {arg_len:d}')

        self.write_byte(arg_len)
        self.write(arg)

    def write_command(self, ccode: int, arg_format: str = None, *args) -> None:
        """Write a Bluetooth command to the stream

        :param ccode: The code of the command
        :param arg_format: The argument format, can be a :func:`struct.pack` string or `varlen` for a variable-length
            argument
        :param args: The arguments to write along with the command, must meet the requirements of the format string
        """
        self.write_byte(ccode)

        if arg_format is not None:
            if arg_format == "varlen":
                self.write_varlen(args[0])
            else:
                self.write_packed(arg_format, *args)

    def read_ack(self) -> None:
        """Read and assert that the next byte in the stream is an acknowledgment

        :raises ValueError: If the byte is not an acknowledgment

        """
        r = self.read_byte()
        if r != ACK_COMMAND_PROCESSED:
            raise ValueError('Byte received is no acknowledgment')

    def read_response(self, rcode: Union[int, Tuple[int, ...], bytes], arg_format: str = None) -> any:
        """Read a Bluetooth command response from the stream

        :param rcode: The expected response code. Can be an int for a single-byte response code,
            or a tuple of ints or a bytes instance for a multi-byte response code
        :param arg_format: The format string to use when decoding the response arguments. Can be None, a
            :func:`struct.unpack` string or `varlen`. If None, no arguments will be read.
        :raises RuntimeError: If the response code is incorrect
        :return: The arguments of the response or () if the response has no arguments
        """
        rcode = resp_code_to_bytes(rcode)

        actual_rcode = self.read(len(rcode))
        if rcode != actual_rcode:
            raise RuntimeError(f'Received incorrect response code: {fmt_hex(rcode)} != {fmt_hex(actual_rcode)}')

        if arg_format is not None:
            if arg_format == "varlen":
                return self.read_varlen()
            else:
                return self.read_packed(arg_format)
        return ()



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import struct
from enum import Enum, auto, unique
from typing import Dict, List, Iterable

from pyshimmer.util import raise_to_next_pow, unpack, flatten_list, bit_is_set


class ChannelDataType:
    """Represents the binary data type and format of a Shimmer data channel

    Every channel that is recorded by a Shimmer device has a specific data type. This class represents the data type
    of a single such channel, and is capable of decoding binary data into the appropriate form.
    """

    def __init__(self, size: int, signed: bool = True, le: bool = True):
        self._size = size
        self._signed = signed
        self._le = le

        self._valid_size = raise_to_next_pow(self.size)
        self._needs_extend = size != self._valid_size

        self._struct_dtypes = {
            1: 'B',
            2: 'H',
            4: 'I',
            8: 'Q',
        }

    @property
    def little_endian(self) -> bool:
        return self._le

    @property
    def big_endian(self) -> bool:
        return not self._le

    @property
    def signed(self) -> bool:
        return self._signed

    @property
    def size(self) -> int:
        return self._size

    def _get_msb(self, val: bytes):
        if self.little_endian:
            return val[-1]
        else:
            return val[0]

    def _get_extension_value(self, val: bytes):
        msb = self._get_msb(val)
        is_negative = (msb >> 7) & 1 == 1

        if self.signed and is_negative:
            return b'\xFF'
        return b'\x00'

    def _extend_value(self, val: bytes) -> bytes:
        ext_value = self._get_extension_value(val)
        suffix = ext_value * (self._valid_size - self.size)

        if self.little_endian:
            return val + suffix
        else:
            return suffix + val

    def _truncate_value(self, val: bytes) -> bytes:
        if self.little_endian:
            return val[:self._size]
        else:
            return val[self._valid_size - self._size:]

    def _get_struct_format(self) -> str:
        stype = self._struct_dtypes[self._valid_size]
        if self.signed:
            stype = stype.lower()

        if self.little_endian:
            prefix = '<'
        else:
            prefix = '>'

        return prefix + stype

    def decode(self, val_bin: bytes) -> any:
        if self._needs_extend:
            val_bin = self._extend_value(val_bin)

        struct_format = self._get_struct_format()
        r_tpl = struct.unpack(struct_format, val_bin)
        return unpack(r_tpl)

    def encode(self, val: int) -> bytes:
        struct_format = self._get_struct_format()
        val_packed = struct.pack(struct_format, val)

        if self._needs_extend:
            return self._truncate_value(val_packed)

        return val_packed


@unique
class EChannelType(Enum):
    """
    Represents the content type of a single data channel recorded by a Shimmer device
    """
    ACCEL_LN_X = auto()
    ACCEL_LN_Y = auto()
    ACCEL_LN_Z = auto()
    VBATT = auto()
    ACCEL_LSM303DLHC_X = auto()
    ACCEL_LSM303DLHC_Y = auto()
    ACCEL_LSM303DLHC_Z = auto()
    MAG_LSM303DLHC_X = auto()
    MAG_LSM303DLHC_Y = auto()
    MAG_LSM303DLHC_Z = auto()
    GYRO_MPU9150_X = auto()
    GYRO_MPU9150_Y = auto()
    GYRO_MPU9150_Z = auto()
    EXTERNAL_ADC_7 = auto()
    EXTERNAL_ADC_6 = auto()
    EXTERNAL_ADC_15 = auto()
    INTERNAL_ADC_1 = auto()
    INTERNAL_ADC_12 = auto()
    INTERNAL_ADC_13 = auto()
    INTERNAL_ADC_14 = auto()
    ACCEL_MPU9150_X = auto()
    ACCEL_MPU9150_Y = auto()
    ACCEL_MPU9150_Z = auto()
    MAG_MPU9150_X = auto()
    MAG_MPU9150_Y = auto()
    MAG_MPU9150_Z = auto()
    TEMP_BMPX80 = auto()
    PRESSURE_BMPX80 = auto()
    GSR_RAW = auto()
    EXG_ADS1292R_1_STATUS = auto()
    EXG_ADS1292R_1_CH1_24BIT = auto()
    EXG_ADS1292R_1_CH2_24BIT = auto()
    EXG_ADS1292R_2_STATUS = auto()
    EXG_ADS1292R_2_CH1_24BIT = auto()
    EXG_ADS1292R_2_CH2_24BIT = auto()
    EXG_ADS1292R_1_CH1_16BIT = auto()
    EXG_ADS1292R_1_CH2_16BIT = auto()
    EXG_ADS1292R_2_CH1_16BIT = auto()
    EXG_ADS1292R_2_CH2_16BIT = auto()
    STRAIN_HIGH = auto()
    STRAIN_LOW = auto()

    TIMESTAMP = auto()


@unique
class ESensorGroup(Enum):
    """
    Represents a sensor of the Shimmer device that can be enabled/disabled via the Bluetooth/Consensys/... API.
    Since one sensor can record more than one channel, there is a one-to-many mapping between sensor and channels.
    """
    # Low-noise accelerometer chip KXRB5-2042
    ACCEL_LN = auto()
    # Battery sensor
    BATTERY = auto()
    # External ADC channel 7
    CH_A7 = auto()
    # External ADC channel 6
    CH_A6 = auto()
    # External ADC channel 15
    CH_A15 = auto()
    # Internal ADC channel 12
    CH_A12 = auto()
    # Internal ADC channel 13, shares the ADC converter with STRAIN
    CH_A13 = auto()
    # Internal ADC channel 14, shares the ADC converter with STRAIN
    CH_A14 = auto()
    # Strain sensor with two channels LOW/HIGH, shares the ADC converter with A13, A14
    STRAIN = auto()
    # Internal ADC channel 1, shares its ADC converter with the GSR sensor
    CH_A1 = auto()
    # GSR sensor, shares the ADC with channel A1
    GSR = auto()
    # MPU9150 Gyro Sensor
    GYRO = auto()
    # Digital accelerometer on the LSM303DLHC chip
    ACCEL_WR = auto()
    # Mag sensor on the LSM303DLHC chip
    MAG = auto()
    # Accelerometer on the MPU9150 chip
    ACCEL_MPU = auto()
    # Mag sensor on the MPU9150 chip
    MAG_MPU = auto()
    # Temperature sensor on the MPU9150 chip, not yet available as channel in the LogAndStream firmware
    TEMP = auto()
    # Pressure sensor on the BMPX80 chip
    PRESSURE = auto()
    # 24 bit channels of the first ADS1292R chip, conflicts with the corresponding 16 bit channel
    EXG1_24BIT = auto()
    # 16 bit channels of the first ADS1292R chip, conflicts with the corresponding 24 bit channel
    EXG1_16BIT = auto()
    # 24 bit channels of the second ADS1292R chip, conflicts with the corresponding 16 bit channel
    EXG2_24BIT = auto()
    # 16 bit channels of the second ADS1292R chip, conflicts with the corresponding 24 bit channel
    EXG2_16BIT = auto()


"""
Assigns each channel type its appropriate data type.
"""
ChDataTypeAssignment: Dict[EChannelType, ChannelDataType] = {
    EChannelType.ACCEL_LN_X: ChannelDataType(2, signed=True, le=True),
    EChannelType.ACCEL_LN_Y: ChannelDataType(2, signed=True, le=True),
    EChannelType.ACCEL_LN_Z: ChannelDataType(2, signed=True, le=True),
    EChannelType.VBATT: ChannelDataType(2, signed=True, le=True),
    EChannelType.ACCEL_LSM303DLHC_X: ChannelDataType(2, signed=True, le=True),
    EChannelType.ACCEL_LSM303DLHC_Y: ChannelDataType(2, signed=True, le=True),
    EChannelType.ACCEL_LSM303DLHC_Z: ChannelDataType(2, signed=True, le=True),
    EChannelType.MAG_LSM303DLHC_X: ChannelDataType(2, signed=True, le=True),
    EChannelType.MAG_LSM303DLHC_Y: ChannelDataType(2, signed=True, le=True),
    EChannelType.MAG_LSM303DLHC_Z: ChannelDataType(2, signed=True, le=True),
    EChannelType.GYRO_MPU9150_X: ChannelDataType(2, signed=True, le=False),
    EChannelType.GYRO_MPU9150_Y: ChannelDataType(2, signed=True, le=False),
    EChannelType.GYRO_MPU9150_Z: ChannelDataType(2, signed=True, le=False),
    EChannelType.EXTERNAL_ADC_7: ChannelDataType(2, signed=False, le=True),
    EChannelType.EXTERNAL_ADC_6: ChannelDataType(2, signed=False, le=True),
    EChannelType.EXTERNAL_ADC_15: ChannelDataType(2, signed=False, le=True),
    EChannelType.INTERNAL_ADC_1: ChannelDataType(2, signed=False, le=True),
    EChannelType.INTERNAL_ADC_12: ChannelDataType(2, signed=False, le=True),
    EChannelType.INTERNAL_ADC_13: ChannelDataType(2, signed=False, le=True),
    EChannelType.INTERNAL_ADC_14: ChannelDataType(2, signed=False, le=True),
    EChannelType.ACCEL_MPU9150_X: None,
    EChannelType.ACCEL_MPU9150_Y: None,
    EChannelType.ACCEL_MPU9150_Z: None,
    EChannelType.MAG_MPU9150_X: None,
    EChannelType.MAG_MPU9150_Y: None,
    EChannelType.MAG_MPU9150_Z: None,
    EChannelType.TEMP_BMPX80: ChannelDataType(2, signed=False, le=False),
    EChannelType.PRESSURE_BMPX80: ChannelDataType(3, signed=False, le=False),
    EChannelType.GSR_RAW: ChannelDataType(2, signed=False, le=True),
    EChannelType.EXG_ADS1292R_1_STATUS: ChannelDataType(1, signed=False, le=True),
    EChannelType.EXG_ADS1292R_1_CH1_24BIT: ChannelDataType(3, signed=True, le=False),
    EChannelType.EXG_ADS1292R_1_CH2_24BIT: ChannelDataType(3, signed=True, le=False),
    EChannelType.EXG_ADS1292R_2_STATUS: ChannelDataType(1, signed=False, le=True),
    EChannelType.EXG_ADS1292R_2_CH1_24BIT: ChannelDataType(3, signed=True, le=False),
    EChannelType.EXG_ADS1292R_2_CH2_24BIT: ChannelDataType(3, signed=True, le=False),
    EChannelType.EXG_ADS1292R_1_CH1_16BIT: ChannelDataType(2, signed=True, le=False),
    EChannelType.EXG_ADS1292R_1_CH2_16BIT: ChannelDataType(2, signed=True, le=False),
    EChannelType.EXG_ADS1292R_2_CH1_16BIT: ChannelDataType(2, signed=True, le=False),
    EChannelType.EXG_ADS1292R_2_CH2_16BIT: ChannelDataType(2, signed=True, le=False),
    EChannelType.STRAIN_HIGH: ChannelDataType(2, signed=False, le=True),
    EChannelType.STRAIN_LOW: ChannelDataType(2, signed=False, le=True),
    EChannelType.TIMESTAMP: ChannelDataType(3, signed=False, le=True)
}

"""
This dictionary contains the mapping from sensor to data channels. Since one sensor can record on multiple channels,
the mapping is one-to-many.
"""
SensorChannelAssignment: Dict[ESensorGroup, List[EChannelType]] = {
    ESensorGroup.ACCEL_LN: [EChannelType.ACCEL_LN_X, EChannelType.ACCEL_LN_Y, EChannelType.ACCEL_LN_Z],
    ESensorGroup.BATTERY: [EChannelType.VBATT],
    ESensorGroup.CH_A7: [EChannelType.EXTERNAL_ADC_7],
    ESensorGroup.CH_A6: [EChannelType.EXTERNAL_ADC_6],
    ESensorGroup.CH_A15: [EChannelType.EXTERNAL_ADC_15],
    ESensorGroup.CH_A12: [EChannelType.INTERNAL_ADC_12],
    ESensorGroup.CH_A13: [EChannelType.INTERNAL_ADC_13],
    ESensorGroup.CH_A14: [EChannelType.INTERNAL_ADC_14],
    ESensorGroup.STRAIN: [EChannelType.STRAIN_HIGH, EChannelType.STRAIN_LOW],
    ESensorGroup.CH_A1: [EChannelType.INTERNAL_ADC_1],
    ESensorGroup.GSR: [EChannelType.GSR_RAW],
    ESensorGroup.GYRO: [EChannelType.GYRO_MPU9150_X, EChannelType.GYRO_MPU9150_Y, EChannelType.GYRO_MPU9150_Z],
    ESensorGroup.ACCEL_WR: [EChannelType.ACCEL_LSM303DLHC_X, EChannelType.ACCEL_LSM303DLHC_Y,
                            EChannelType.ACCEL_LSM303DLHC_Z],
    ESensorGroup.MAG: [EChannelType.MAG_LSM303DLHC_X, EChannelType.MAG_LSM303DLHC_Y, EChannelType.MAG_LSM303DLHC_Z],
    ESensorGroup.ACCEL_MPU: [EChannelType.ACCEL_MPU9150_X, EChannelType.ACCEL_MPU9150_Y, EChannelType.ACCEL_MPU9150_Z],
    ESensorGroup.MAG_MPU: [EChannelType.MAG_MPU9150_X, EChannelType.MAG_MPU9150_Y, EChannelType.MAG_MPU9150_Z],
    ESensorGroup.PRESSURE: [EChannelType.TEMP_BMPX80, EChannelType.PRESSURE_BMPX80],
    ESensorGroup.EXG1_24BIT: [EChannelType.EXG_ADS1292R_1_STATUS, EChannelType.EXG_ADS1292R_1_CH1_24BIT,
                              EChannelType.EXG_ADS1292R_1_CH2_24BIT],
    ESensorGroup.EXG1_16BIT: [EChannelType.EXG_ADS1292R_1_STATUS, EChannelType.EXG_ADS1292R_1_CH1_16BIT,
                              EChannelType.EXG_ADS1292R_1_CH2_16BIT],
    ESensorGroup.EXG2_24BIT: [EChannelType.EXG_ADS1292R_2_STATUS, EChannelType.EXG_ADS1292R_2_CH1_24BIT,
                              EChannelType.EXG_ADS1292R_2_CH2_24BIT],
    ESensorGroup.EXG2_16BIT: [EChannelType.EXG_ADS1292R_2_STATUS, EChannelType.EXG_ADS1292R_2_CH1_16BIT,
                              EChannelType.EXG_ADS1292R_2_CH2_16BIT],

    # The MPU9150 Temp sensor is not yet available as a channel in the LogAndStream firmware
    ESensorGroup.TEMP: [],
}

"""
The sensors are enabled via a multi-byte bitfield that currently stretches a total of three bytes. This dictionary
contains the bitfield position for every sensor in this bitfield.
"""
SensorBitAssignments: Dict[ESensorGroup, int] = {
    ESensorGroup.ACCEL_LN: 0x80 << 0 * 8,
    ESensorGroup.GYRO: 0x40 << 0 * 8,
    ESensorGroup.MAG: 0x20 << 0 * 8,
    ESensorGroup.EXG1_24BIT: 0x10 << 0 * 8,
    ESensorGroup.EXG2_24BIT: 0x08 << 0 * 8,
    ESensorGroup.GSR: 0x04 << 0 * 8,
    ESensorGroup.CH_A7: 0x02 << 0 * 8,
    ESensorGroup.CH_A6: 0x01 << 0 * 8,

    ESensorGroup.STRAIN: 0x80 << 1 * 8,
    # No assignment             0x40 << 1 * 8,
    ESensorGroup.BATTERY: 0x20 << 1 * 8,
    ESensorGroup.ACCEL_WR: 0x10 << 1 * 8,
    ESensorGroup.CH_A15: 0x08 << 1 * 8,
    ESensorGroup.CH_A1: 0x04 << 1 * 8,
    ESensorGroup.CH_A12: 0x02 << 1 * 8,
    ESensorGroup.CH_A13: 0x01 << 1 * 8,

    ESensorGroup.CH_A14: 0x80 << 2 * 8,
    ESensorGroup.ACCEL_MPU: 0x40 << 2 * 8,
    ESensorGroup.MAG_MPU: 0x20 << 2 * 8,
    ESensorGroup.EXG1_16BIT: 0x10 << 2 * 8,
    ESensorGroup.EXG2_16BIT: 0x08 << 2 * 8,
    ESensorGroup.PRESSURE: 0x04 << 2 * 8,
    ESensorGroup.TEMP: 0x02 << 2 * 8,
}

SensorOrder: Dict[ESensorGroup, int] = {
    ESensorGroup.ACCEL_LN: 1,
    ESensorGroup.BATTERY: 2,
    ESensorGroup.CH_A7: 3,
    ESensorGroup.CH_A6: 4,
    ESensorGroup.CH_A15: 5,
    ESensorGroup.CH_A12: 6,
    ESensorGroup.CH_A13: 7,
    ESensorGroup.CH_A14: 8,
    ESensorGroup.STRAIN: 9,
    ESensorGroup.CH_A1: 10,
    ESensorGroup.GSR: 11,
    ESensorGroup.GYRO: 12,
    ESensorGroup.ACCEL_WR: 13,
    ESensorGroup.MAG: 14,
    ESensorGroup.ACCEL_MPU: 15,
    ESensorGroup.MAG_MPU: 16,
    ESensorGroup.PRESSURE: 17,
    ESensorGroup.EXG1_24BIT: 18,
    ESensorGroup.EXG1_16BIT: 19,
    ESensorGroup.EXG2_24BIT: 20,
    ESensorGroup.EXG2_16BIT: 21,
}

ENABLED_SENSORS_LEN = 0x03
SENSOR_DTYPE = ChannelDataType(size=ENABLED_SENSORS_LEN, signed=False, le=True)


def get_enabled_channels(sensors: List[ESensorGroup]) -> List[EChannelType]:
    """Determine the set of data channels for a set of enabled sensors

    There exists a one-to-many mapping between enabled sensors and their corresponding data channels. This function
    determines the set of necessary channels for a given set of enabled sensors.

    Args:
        sensors: A list of sensors that are enabled on a Shimmer

    Returns:
        A list of channels in the corresponding order

    """
    channels = [SensorChannelAssignment[e] for e in sensors]
    return flatten_list(channels)


def get_ch_dtypes(channels: List[EChannelType]) -> List[ChannelDataType]:
    """Return the channel data types for a set of channels

    Args:
        channels: A list of channels
    Returns:
        A list of channel data types with the same order
    """
    dtypes = [ChDataTypeAssignment[ch] for ch in channels]
    return dtypes


def sensors2bitfield(sensors: Iterable[ESensorGroup]) -> int:
    """Convert an iterable of sensors into the corresponding bitfield transmitted to the Shimmer

    :param sensors: A list of active sensors
    :return: A bitfield that conveys the set of active sensors to the Shimmer
    """
    bitfield = 0
    for sensor in sensors:
        bit_pos = SensorBitAssignments[sensor]
        bitfield |= bit_pos

    return bitfield


def serialize_sensorlist(sensors: Iterable[ESensorGroup]) -> bytes:
    """Serialize a list of sensors to the three-byte bitfield accepted by the Shimmer

    :param sensors: The list of sensors
    :return: A byte string with length 3 that encodes the sensors
    """
    bitfield = sensors2bitfield(sensors)
    return SENSOR_DTYPE.encode(bitfield)


def bitfield2sensors(bitfield: int) -> List[ESensorGroup]:
    """Decode a bitfield returned from the Shimmer to a list of active sensors

    :param bitfield: The bitfield received from the Shimmer encoding the active sensors
    :return: The corresponding list of active sensors
    """
    enabled_sensors = []
    for sensor in ESensorGroup:
        bit_pos = SensorBitAssignments[sensor]
        if bit_is_set(bitfield, bit_pos):
            enabled_sensors += [sensor]

    return sort_sensors(enabled_sensors)


def deserialize_sensors(bitfield_bin: bytes) -> List[ESensorGroup]:
    """Deserialize the list of active sensors from the three-byte input received from the Shimmer

    :param bitfield_bin: The input bitfield as byte string with length 3
    :return: The list of active sensors
    """
    bitfield = SENSOR_DTYPE.decode(bitfield_bin)
    return bitfield2sensors(bitfield)


def sort_sensors(sensors: Iterable[ESensorGroup]) -> List[ESensorGroup]:
    """Sorts the sensors in the list according to the sensor order

    This function is useful to determine the order in which sensor data will appear in a data file by ordering
    the list of sensors according to their order in the file.

    Args:
        sensors: An unsorted list of sensors

    Returns:
        A list with the same sensors as content but sorted according to their appearance order in the data file

    """

    def sort_key_fn(x):
        return SensorOrder[x]

    sensors_sorted = sorted(sensors, key=sort_key_fn)
    return sensors_sorted



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
UART_SET = 0x01
UART_GET = 0x03

UART_RESPONSE = 0x02
UART_ACK_RESPONSE = 0xFF

UART_BAD_CMD_RESPONSE = 0xFc
UART_BAD_ARG_RESPONSE = 0xFD
UART_BAD_CRC_RESPONSE = 0xFE

CRC_INIT = 0xB0CA
START_CHAR = 0x24

UART_COMP_SHIMMER = 0x01
UART_COMP_BAT = 0x02
UART_COMP_DAUGHTER_CARD = 0x03
UART_COMP_D_ACCEL = 0x04
UART_COMP_GSR = 0x05

UART_PROP_ENABLE = 0x00
UART_PROP_SAMPLE_RATE = 0x01
UART_PROP_MAC = 0x02
UART_PROP_VER = 0x03
UART_PROP_RWC_CFG_TIME = 0x04
UART_PROP_CURR_LOCAL_TIME = 0x05
UART_PROP_INFOMEM = 0x06
UART_PROP_CARD_ID = 0x02

UART_INFOMEM_EXG_OFFSET = 0x0A



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import re
from enum import Enum
from typing import List, Dict, Tuple

from pyshimmer.util import bit_is_set, fmt_hex
from .channels import EChannelType


class ExGMux(Enum):
    NORMAL = 0x00
    SHORTED = 0x01
    RLD_MEASURE = 0x02
    MVDD = 0x03
    TEMP_SENSOR = 0x04
    TEST_SIGNAL = 0x05
    RLD_DRP = 0x06
    RLD_DRM = 0x07
    RLD_DRPM = 0x08
    INPUT3 = 0x09
    RESERVED = 0x0A


class ExGRLDLead(Enum):
    RLD1P = 0x01 << 0
    RLD1N = 0x01 << 1
    RLD2P = 0x01 << 2
    RLD2N = 0x01 << 3


class ERLDRef(Enum):
    EXTERNAL = 0x00
    INTERNAL = 0x01


class ExGRegister:
    GAIN_MAP = {
        0: 6, 1: 1, 2: 2, 3: 3, 4: 4, 5: 8, 6: 12,
    }
    DATA_RATE_MAP = {
        0: 125, 1: 250, 2: 500, 3: 1000, 4: 2000, 5: 4000, 6: 8000, 7: -1
    }
    PD_BIT = 0x01 << 7
    RLD_PD_BIT = 0x01 << 5

    def __init__(self, reg_bin: bytes):
        if len(reg_bin) < 10:
            raise ValueError('Binary register content must have length 10')

        self._reg_bin = reg_bin

    def __str__(self) -> str:
        def print_ch(ch_id: int) -> str:
            return f'Channel {ch_id + 1:2d}\n' + \
                   f'\tPowerdown: {self.is_ch_powerdown(ch_id)}\n' + \
                   f'\tGain: {self.get_ch_gain(ch_id):2d}\n' + \
                   f'\tMultiplexer: {self.get_ch_mux(ch_id).name} ({self.get_ch_mux_bin(ch_id):#06b})\n'

        def fmt_rld_channels(ch_names) -> str:
            ch_names = [ch.name for ch in ch_names] if len(ch_names) > 0 else ['None']
            return ', '.join(ch_names)

        reg_bin_str = fmt_hex(self._reg_bin)
        obj_str = f'ExGRegister:\n' + \
                  f'Data Rate: {self.data_rate:4d}\n' + \
                  f'RLD Powerdown: {self.rld_powerdown}\n' + \
                  f'RLD Channels: {fmt_rld_channels(self.rld_channels)}\n' + \
                  f'RLD Reference: {self.rld_ref.name}\n' + \
                  f'Binary: {reg_bin_str}\n'
        obj_str += print_ch(0)
        obj_str += print_ch(1)
        return obj_str

    def __eq__(self, other: "ExGRegister") -> bool:
        return self._reg_bin == other._reg_bin

    @staticmethod
    def check_ch_id(ch_id: int) -> None:
        if not 0 <= ch_id <= 1:
            raise ValueError('Channel ID must be 0 or 1')

    def _get_ch_byte(self, ch_id: int) -> int:
        ch_offset = 3 + ch_id
        return self._reg_bin[ch_offset]

    def _get_rld_byte(self) -> int:
        return self._reg_bin[0x05]

    def get_ch_gain(self, ch_id: int) -> int:
        self.check_ch_id(ch_id)

        ch_byte = self._get_ch_byte(ch_id)
        gain_bin = (ch_byte & 0x70) >> 4

        return self.GAIN_MAP[gain_bin]

    def get_ch_mux_bin(self, ch_id: int) -> int:
        self.check_ch_id(ch_id)

        ch_byte = self._get_ch_byte(ch_id)
        return ch_byte & 0x0F

    def get_ch_mux(self, ch_id: int) -> ExGMux:
        return ExGMux(self.get_ch_mux_bin(ch_id))

    def is_ch_powerdown(self, ch_id: int) -> bool:
        self.check_ch_id(ch_id)

        ch_byte = self._get_ch_byte(ch_id)
        return bit_is_set(ch_byte, self.PD_BIT)

    @property
    def binary(self):
        return self._reg_bin

    @property
    def ch1_gain(self) -> int:
        return self.get_ch_gain(0)

    @property
    def ch2_gain(self) -> int:
        return self.get_ch_gain(1)

    @property
    def ch1_mux(self) -> ExGMux:
        return self.get_ch_mux(0)

    @property
    def ch2_mux(self) -> ExGMux:
        return self.get_ch_mux(1)

    @property
    def ch1_powerdown(self) -> bool:
        return self.is_ch_powerdown(0)

    @property
    def ch2_powerdown(self) -> bool:
        return self.is_ch_powerdown(1)

    @property
    def data_rate(self) -> int:
        dr_bin = self._reg_bin[0] & 0x07
        return self.DATA_RATE_MAP[dr_bin]

    @property
    def rld_powerdown(self) -> bool:
        rld_byte = self._get_rld_byte()
        return not bit_is_set(rld_byte, self.RLD_PD_BIT)

    @property
    def rld_channels(self) -> List[ExGRLDLead]:
        rld_byte = self._get_rld_byte()
        return [ch for ch in ExGRLDLead if bit_is_set(rld_byte, ch.value)]

    @property
    def rld_ref(self) -> ERLDRef:
        ref_byte = self._reg_bin[9]
        rld_ref = (ref_byte >> 1) & 0x01
        return ERLDRef(rld_ref)


ExG_ChType_Chip_Assignment: Dict[EChannelType, Tuple[int, int]] = {
    EChannelType.EXG_ADS1292R_1_CH1_24BIT: (0, 0),
    EChannelType.EXG_ADS1292R_1_CH1_16BIT: (0, 0),

    EChannelType.EXG_ADS1292R_1_CH2_24BIT: (0, 1),
    EChannelType.EXG_ADS1292R_1_CH2_16BIT: (0, 1),

    EChannelType.EXG_ADS1292R_2_CH1_24BIT: (1, 0),
    EChannelType.EXG_ADS1292R_2_CH1_16BIT: (1, 0),

    EChannelType.EXG_ADS1292R_2_CH2_24BIT: (1, 1),
    EChannelType.EXG_ADS1292R_2_CH2_16BIT: (1, 1),
}


def is_exg_ch(ch_type: EChannelType) -> bool:
    """
    Returns true if the signal that this channel type describes was recorded by a ExG chip

    Args:
        ch_type: The EChannelType of the signal

    Returns:
        True if the channel type belongs to the ExG chips, otherwise False
    """
    # This is hacky but protected by unit tests
    regex = re.compile(r'EXG_ADS1292R_\d_CH\d_\d{2}BIT')
    return regex.match(ch_type.name) is not None


def get_exg_ch(ch_type: EChannelType) -> Tuple[int, int]:
    """
    Each ExG Chip EChannelType originates from a specific ExG chip and channel. This function returns a tuple that
    specifices which chip and channel a certain signal EChannelType was recorded with.

    Args:
        ch_type: The EChannelType of the signal

    Returns:
        A tuple of ints which represents the chip id {0, 1} and the channel id {0, 1}.

    """
    return ExG_ChType_Chip_Assignment[ch_type]



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2023  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from unittest import TestCase

from pyshimmer.dev.base import sr2dr, dr2sr, sec2ticks, ticks2sec


class DeviceBaseTest(TestCase):

    def test_sr2dr(self):
        r = sr2dr(1024.0)
        self.assertEqual(r, 32)

        r = sr2dr(500.0)
        self.assertEqual(r, 66)

    def test_dr2sr(self):
        r = dr2sr(65)
        self.assertEqual(round(r), 504)

        r = dr2sr(32)
        self.assertEqual(r, 1024.0)

        r = dr2sr(64)
        self.assertEqual(r, 512.0)

    def test_sec2ticks(self):
        r = sec2ticks(1.0)
        self.assertEqual(r, 32768)

    def test_ticks2sec(self):
        r = ticks2sec(32768)
        self.assertEqual(r, 1.0)

        r = ticks2sec(65536)
        self.assertEqual(r, 2.0)



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pyshimmer.dev.channels import ESensorGroup

SR_OFFSET = 0x00

ENABLED_SENSORS_OFFSET = 0x03

RTC_CLOCK_DIFF_OFFSET = 0x2C

START_TS_OFFSET = 0xFB
START_TS_LEN = 0x5

TRIAL_CONFIG_OFFSET = 0x10

DATA_LOG_OFFSET = 0x100
BLOCK_LEN = 0x200

TRIAL_CONFIG_SYNC = 0x04 << 8 * 0
TRIAL_CONFIG_MASTER = 0x02 << 8 * 0

EXG_REG_OFFSET = 0x38
EXG_REG_LEN = 0x0A

# The file offsets at which the calibration parameters of the respective sensor can be found
TRIAXCAL_FILE_OFFSET = {
    ESensorGroup.ACCEL_LN: 0x8B,
    ESensorGroup.ACCEL_WR: 0x4C,
    ESensorGroup.GYRO: 0x61,
    ESensorGroup.MAG: 0x76,
}

# Scaling value by which the calibration offset will be scaled upon deserialization
TRIAXCAL_OFFSET_SCALING = {
    ESensorGroup.ACCEL_LN: 1.0,
    ESensorGroup.ACCEL_WR: 1.0,
    ESensorGroup.GYRO: 1.0,
    ESensorGroup.MAG: 1.0,
}

# Scaling value by which the calibration gain will be scaled upon deserialization
TRIAXCAL_GAIN_SCALING = {
    ESensorGroup.ACCEL_LN: 1.0,
    ESensorGroup.ACCEL_WR: 1.0,
    ESensorGroup.GYRO: 1.0 / 100.0,
    ESensorGroup.MAG: 1.0,
}

# Scaling value by which the calibration alignment matrix will be scaled upon deserialization
TRIAXCAL_ALIGNMENT_SCALING = {
    ESensorGroup.ACCEL_LN: 1.0 / 100.0,
    ESensorGroup.ACCEL_WR: 1.0 / 100.0,
    ESensorGroup.GYRO: 1.0 / 100.0,
    ESensorGroup.MAG: 1.0 / 100.0,
}

TRIAXCAL_SENSORS = list(TRIAXCAL_FILE_OFFSET.keys())

EXG_ADC_OFFSET = 0.0
EXG_ADC_REF_VOLT = 2.42  # Volts



# pyshimmer - API for Shimmer sensor devices
# Copyright (C) 2020  Lukas Magel

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

from abc import ABC, abstractmethod
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Dict, List, BinaryIO

import numpy as np

from pyshimmer.dev.base import ticks2sec, dr2sr
from pyshimmer.dev.channels import ChDataTypeAssignment, get_enabled_channels, EChannelType
from pyshimmer.dev.exg import is_exg_ch, get_exg_ch, ExGRegister
from pyshimmer.reader.binary_reader import ShimmerBinaryReader
from pyshimmer.reader.reader_const import EXG_ADC_REF_VOLT, EXG_ADC_OFFSET, TRIAXCAL_SENSORS
from pyshimmer.util import unwrap


def unwrap_device_timestamps(ts_dev: np.ndarray) -> np.ndarray:
    ts_dtype = ChDataTypeAssignment[EChannelType.TIMESTAMP]
    return unwrap(ts_dev, 2 ** (8 * ts_dtype.size))


def fit_linear_1d(xp, fp, x):
    fit_coef = np.polyfit(xp, fp, 1)
    fn = np.poly1d(fit_coef)
    return fn(x)


class ChannelPostProcessor(ABC):

    @abstractmethod
    def process(self, channels: Dict[EChannelType, np.ndarray], reader: ShimmerBinaryReader) -> \
            Dict[EChannelType, np.ndarray]:
        pass


class SingleChannelProcessor(ChannelPostProcessor, ABC):

    def __init__(self, ch_types: List[EChannelType] = None):
        super().__init__()
        self._ch_types = ch_types

    def process(self, channels: Dict[EChannelType, np.ndarray], reader: ShimmerBinaryReader) -> \
            Dict[EChannelType, np.ndarray]:

        if self._ch_types is None:
            ch_types = list(channels.keys())
        else:
            ch_types = [t for t in self._ch_types if t in channels]

        result = channels.copy()
        for ch_type in ch_types:
            result[ch_type] = self.process_channel(ch_type, channels[ch_type], reader)

        return result

    @abstractmethod
    def process_channel(self, ch_type: EChannelType, y: np.ndarray, reader: ShimmerBinaryReader) -> np.ndarray:
        pass


class ExGProcessor(SingleChannelProcessor):

    def __init__(self):
        exg_channels = [t for t in EChannelType if is_exg_ch(t)]
        super().__init__(exg_channels)

    def process_channel(self, ch_type: EChannelType, y: np.ndarray, reader: ShimmerBinaryReader) -> np.ndarray:
        chip_id, ch_id = get_exg_ch(ch_type)
        exg_reg = reader.get_exg_reg(chip_id)
        gain = exg_reg.get_ch_gain(ch_id)

        ch_dtype = ChDataTypeAssignment[ch_type]
        resolution = 8 * ch_dtype.size
        sensitivity = EXG_ADC_REF_VOLT / (2 ** (resolution - 1) - 1)

        # According to formula in Shimmer ECG User Guide
        y_volt = (y - EXG_ADC_OFFSET) * sensitivity / gain
        return y_volt


class PPGProcessor(SingleChannelProcessor):

    def __init__(self):
        super().__init__([EChannelType.INTERNAL_ADC_13])

    def process_channel(self, ch_type: EChannelType, y: np.ndarray, reader: ShimmerBinaryReader) -> np.ndarray:
        # Convert from mV to V
        return y / 1000.0


class TriAxCalProcessor(ChannelPostProcessor):

    def process(self, channels: Dict[EChannelType, np.ndarray], reader: ShimmerBinaryReader) -> \
            Dict[EChannelType, np.ndarray]:
        result = channels.copy()

        active_sensors = [s for s in reader.enabled_sensors if s in TRIAXCAL_SENSORS]
        for sensor in active_sensors:
            sensor_channels = get_enabled_channels([sensor])
            channel_data = np.stack([channels[c] for c in sensor_channels])
            o, g, a = reader.get_triaxcal_params(sensor)

            g_a = np.matmul(g, a)
            r = np.linalg.solve(g_a, channel_data - o[..., None])

            for i, ch in enumerate(sensor_channels):
                result[ch] = r[i]

        return result


class ShimmerReader:

    def __init__(self, fp: BinaryIO = None, bin_reader: ShimmerBinaryReader = None,
                 sync: bool = True, post_process: bool = True, processors: List[ChannelPostProcessor] = None):
        if fp is not None:
            self._bin_reader = ShimmerBinaryReader(fp)
        elif bin_reader is not None:
            self._bin_reader = bin_reader
        else:
            raise ValueError('Need to provide file object or binary reader as parameter.')

        self._ts = None
        self._ch_samples = {}
        self._sync = sync

        self._pp = post_process
        if processors is not None:
            self._processors = processors
        else:
            self._processors = [
                ExGProcessor(),
                PPGProcessor(),
                TriAxCalProcessor(),
            ]

    @staticmethod
    def _apply_synchronization(data_ts: np.ndarray, offset_index: np.ndarray, offsets: np.ndarray):
        # We discard all synchronization offsets for which we do not possess timestamps.
        index_safe = offset_index[offset_index < len(data_ts)]

        offsets_ts = data_ts[index_safe]
        data_offsets = fit_linear_1d(offsets_ts, offsets, data_ts)

        aligned_ts = data_ts - data_offsets
        return aligned_ts

    def _apply_clock_offsets(self, ts: np.ndarray):
        # First, we need calculate absolute timestamps relative to the boot-up time of the Shimmer.
        # In order to do so, we use the 40bit initial timestamp to calculate an offset to apply to
        # each timestamp.
        boot_offset = self._bin_reader.start_timestamp - ts[0]
        ts_boot = ts + boot_offset

        if self._bin_reader.has_global_clock:
            return ts_boot + self._bin_reader.global_clock_diff
        else:
            return ts_boot

    def _process_signals(self, channels: Dict[EChannelType, np.ndarray]) -> Dict[EChannelType, np.ndarray]:
        result = channels.copy()

        for processor in self._processors:
            result = processor.process(result, self._bin_reader)

        return result

    def load_file_data(self):
        samples, sync_offsets = self._bin_reader.read_data()
        ts_raw = samples.pop(EChannelType.TIMESTAMP)

        ts_unwrapped = unwrap_device_timestamps(ts_raw)
        ts_sane = self._apply_clock_offsets(ts_unwrapped)

        if self._sync and self._bin_reader.has_sync:
            ts_sane = self._apply_synchronization(ts_sane, *sync_offsets)

        if self._pp:
            self._ch_samples = self._process_signals(samples)
        else:
            self._ch_samples = samples

        self._ts = ticks2sec(ts_sane)

    def get_exg_reg(self, chip_id: int) -> ExGRegister:
        return self._bin_reader.get_exg_reg(chip_id)

    def __getitem__(self, item: EChannelType) -> np.ndarray:
        if item == EChannelType.TIMESTAMP:
            return self.timestamp

        return self._ch_samples[item]

    @property
    def timestamp(self) -> np.ndarray:
        return self._ts

    @property
    def channels(self) -> List[EChannelType]:
        # We return all but the first channel which are the timestamps
        return self._bin_reader.enabled_channels[1:]

    @property
    def sample_rate(self) -> float:
        return dr2sr(self._bin_reader.sample_rate)

    @property
    def exg_reg1(self) -> ExGRegister:
        return self.get_exg_reg(0)

    @property
    def exg_reg2(self) -> ExGRegister:
        return self.get_exg_reg(1)
